from typing import Optional, Tuple, List

from eth_typing import ChecksumAddress
from web3 import Web3
from web3.contract import Contract

from rubi.contracts.base_contract import BaseContract
from rubi.network import Network


class RubiconMarket(BaseContract):
    """This class represents the RubiconMarket.sol contract and by default has read functionality.
    If a wallet and key are passed in instantiation then this class can also be used to write to the contract instance.

    :param w3: Web3 instance
    :type w3: Web3
    :param contract: Contract instance
    :type contract: Contract
    :param wallet: a wallet address of the signer (optional, default is None)
    :type wallet: Optional[ChecksumAddress]
    :param key: the private key of the signer (optional, default is None)
    :type key: Optional[str]
    """

    def __init__(
        self,
        w3: Web3,
        contract: Contract,
        wallet: Optional[ChecksumAddress] = None,
        key: Optional[str] = None
    ) -> None:
        """constructor method"""
        super().__init__(
            w3=w3,
            contract=contract,
            wallet=wallet,
            key=key
        )

    @classmethod
    def from_network(
        cls,
        network: Network,
        wallet: Optional[ChecksumAddress] = None,
        key: Optional[str] = None
    ) -> "RubiconMarket":
        """Create a RubiconMarket instance based on a Network instance.

        :param network: A Network instance.
        :type network: Network
        :param wallet: Optional wallet address to use for interacting with the contract (optional, default is None).
        :type wallet: Optional[ChecksumAddress]
        :param key: Optional private key for the wallet (optional, default is None).
        :type key: Optional[str]
        :return: A RubiconMarket instance based on the Network instance.
        :rtype: RubiconMarket
        """
        return cls.from_address_and_abi(
            w3=network.w3,
            address=network.rubicon.market.address,
            contract_abi=network.rubicon.market.abi,
            wallet=wallet,
            key=key
        )

    ######################################################################
    # read calls
    ######################################################################

    # makerFee() -> (uint265)
    def get_maker_fee(self) -> int:
        """Returns the maker fee on Rubicon

        :return: the maker fee
        :rtype: int
        """

        return self.contract.functions.makerFee().call()

    # getOffer(id (uint256)) -> (uint256, address, uint256, address)
    def get_offer(self, id: int) -> Tuple[int, ChecksumAddress, int, ChecksumAddress]:
        """Returns the offer associated with the provided id

        :param id: the id of the offer being queried
        :type id: int
        :return: a description of the offer as (pay_amt, pay_gem, buy_amt, buy_gem)
        :rtype: Tuple[int, ChecksumAddress, int, ChecksumAddress]
        """
        return self.contract.functions.getOffer(id).call()

    # getMinSell(pay_gem (address)) -> uint256
    def get_min_sell(self, pay_gem: ChecksumAddress) -> int:
        """Returns the minimum sell amount for an offer

        :param pay_gem: the address of the token being sold by the maker
        :type pay_gem: str
        :return: the minimum amount of pay_gem that can be sold in an offer
        :rtype: int
        """

        return self.contract.functions.getMinSell(pay_gem).call()

    # getBestOffer(sell_gem (address), buy_gem (address)) -> uint256
    def get_best_offer(self, sell_gem: ChecksumAddress, buy_gem: ChecksumAddress) -> int:
        """Returns the best offer for the given pair of tokens

        :param sell_gem: the address of the token being sold by the maker
        :type sell_gem: str
        :param buy_gem: the address of the token being bought by the maker
        :type buy_gem: str
        :return: the id of the best offer on the book, None if there is no offer on the book
        :rtype: int
        """

        return self.contract.functions.getBestOffer(sell_gem, buy_gem).call()

    # getWorseOffer(id (uint256)) -> uint256
    def get_worse_offer(self, id: int) -> int:
        """Returns the id of the offer that is worse than the given offer

        :param id: the id of the offer
        :type id: int
        :return: the id of the offer that is worse than the given offer, none if there is no worse offer
        :rtype: int
        """

        return self.contract.functions.getWorseOffer(id).call()

    # getBetterOffer(id (uint256)) -> uint256
    def get_better_offer(self, id: int) -> int:
        """Returns the id of the offer that is better than the given offer

        :param id: the id of the offer
        :type id: int
        :return: the id of the offer that is better than the given offer, none if there is no better offer
        :rtype: int
        """

        return self.contract.functions.getBetterOffer(id).call()

    # getOfferCount(sell_gem (address), buy_gem (address)) -> uint256
    def get_offer_count(self, sell_gem: ChecksumAddress, buy_gem: ChecksumAddress) -> int:
        """Returns the number of offers for a token pair

        :param sell_gem: the address of the token being sold by the maker
        :type sell_gem: ChecksumAddress
        :param buy_gem: the address of the token being bought by the maker
        :type buy_gem: ChecksumAddress
        :return: the number of offers for a token pair, None if there are no offers for the token pair
        :rtype: int
        """

        return self.contract.functions.getOfferCount(sell_gem, buy_gem).call()

    # calculateFees(amount (uint256), isPay (bool)) -> uint256
    def calculate_fees(self, amount: int) -> int:
        """Calculate fees on an amount

        :param amount: the address of the token being bought
        :type amount: int
        :return: the calculated fees on the amount
        :rtype: int
        """
        return self.contract.functions.calculateFees(amount, True).call()

    # getBuyAmountWithFee(buy_gem (address), pay_gem (address), pay_amt (unit256)) ->
    # (buy_amt (uint256), approvalAmount (uint256))
    def get_buy_amount_with_fee(
        self,
        buy_gem: ChecksumAddress,
        pay_gem: ChecksumAddress,
        pay_amt: int
    ) -> Tuple[int, int]:
        """Returns the amount of the buy_gem you will receive if you send the pay_amt to the contract along with the
        amount to approve for the transaction

        :param buy_gem: the address of the token being bought
        :type buy_gem: ChecksumAddress
        :param pay_gem: the address of the token being sold
        :type pay_gem: ChecksumAddress
        :param pay_amt: the amount of the token being sold to receive the token being bought
        :type pay_amt: int
        :return: (buy_amt, approvalAmount) the amount of tokens that will be received and the amount to approve for the
            transaction
        :rtype: Tuple[int, int]
        """
        return self.contract.functions.getBuyAmountWithFee(buy_gem, pay_gem, pay_amt).call()

    # getPayAmountWithFee(pay_gem (address), buy_gem (address), buy_amt (unit256)) ->
    # (buy_amt (uint256), approvalAmount (uint256))
    def get_pay_amount_with_fee(
        self,
        pay_gem: ChecksumAddress,
        buy_gem: ChecksumAddress,
        buy_amt: int
    ) -> Tuple[int, int]:
        """Returns the amount of the pay_gem you will need to pay to the contract to receive the buy_amt along with the
        amount to approve for the transaction

        :param buy_gem: the address of the token being bought
        :type buy_gem: ChecksumAddress
        :param pay_gem: the address of the token being sold
        :type pay_gem: ChecksumAddress
        :param buy_amt: the amount of the token being bought
        :type buy_amt: int
        :return: (pay_amt, approvalAmount) the amount of tokens that will be paid and the amount to approve for the
            transaction
        :rtype: Tuple[int, int]
        """
        return self.contract.functions.getPayAmountWithFee(buy_gem, pay_gem, buy_amt).call()

    ######################################################################
    # write calls
    ######################################################################

    # offer offer(pay_amt (uint256), pay_gem (address), buy_amt (uint256), buy_gem (address), pos (uint256),
    # owner (address), recipient (address)) -> uint256
    def offer(
        self,
        pay_amt: int,
        pay_gem: ChecksumAddress,
        buy_amt: int,
        buy_gem: ChecksumAddress,
        pos: int = 0,
        rounding: bool = True,
        owner: Optional[ChecksumAddress] = None,
        recipient: Optional[ChecksumAddress] = None,
        nonce: Optional[int] = None,
        gas: int = 350000,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None
    ) -> str:
        """Make a new offer to buy the buy_amt of the buy_gem token in exchange for the pay_amt of the pay_gem token

        :param pay_amt: the amount of the token being sold
        :type pay_amt: int
        :param pay_gem: the address of the token being sold
        :type pay_gem: ChecksumAddress
        :param buy_amt: the amount of the token being bought
        :type buy_amt: int
        :param buy_gem: the address of the token being bought
        :type buy_gem: ChecksumAddress
        :param pos: position of the offer in the linked list, default to 0 unless the maker knows the position they want
            to insert the offer at
        :type pos: int
        :param rounding: add rounding to match "close enough" orders, defaults to True
        :type: rounding: bool
        :param owner: the owner of the offer, defaults to the wallet that was provided in instantiating this class.
            (optional, default is None)
        :type owner: Optional[ChecksumAddress]
        :param recipient: the recipient of the offer's fill, defaults to the wallet that was provided in instantiating
            this class (optional, default is None)
        :type recipient: Optional[ChecksumAddress]
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce.
            (optional, default is None)
        :type nonce: Optional[int]
        :param gas: gas limit of the transaction, defaults to a very high estimate made when writing the class
        :type gas: int
        :param max_fee_per_gas: max fee that can be paid for gas, defaults to
            max_priority_fee (from chain) + (2 * base fee per gas of latest block) (optional, default is None)
        :type max_fee_per_gas: Optional[int]
        :param max_priority_fee_per_gas: max priority fee that can be paid for gas, defaults to calling the chain to
            estimate the max_priority_fee_per_gas (optional, default is None)
        :type max_priority_fee_per_gas: Optional[int]
        :return: transaction hash
        :rtype: str
        """

        if not self.signing_permissions:
            raise Exception(f"cannot write transaction without signing rights. "
                            f"re-instantiate {self.__class__} with a wallet and private key")

        owner = owner if owner is not None else self.wallet
        recipient = recipient if recipient is not None else self.wallet

        offer = self.contract.functions.offer(pay_amt, pay_gem, buy_amt, buy_gem, pos, rounding, owner, recipient)

        return self._default_transaction_handler(
            instantiated_contract_function=offer,
            gas=gas,
            nonce=nonce,
            max_fee_per_gas=max_fee_per_gas,
            max_priority_fee_per_gas=max_priority_fee_per_gas
        )

    # cancel(id (uint256)) -> bool
    def cancel(
        self,
        id: int,
        nonce: Optional[int] = None,
        gas: int = 350000,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None
    ) -> str:
        """Cancel an offer by offer id

        :param id: the id of the offer to cancel
        :type id: int
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce.
            (optional, default is None)
        :type nonce: Optional[int]
        :param gas: gas limit of the transaction, defaults to a very high estimate made when writing the class
        :type gas: int
        :param max_fee_per_gas: max fee that can be paid for gas, defaults to
            max_priority_fee (from chain) + (2 * base fee per gas of latest block) (optional, default is None)
        :type max_fee_per_gas: Optional[int]
        :param max_priority_fee_per_gas: max priority fee that can be paid for gas, defaults to calling the chain to
            estimate the max_priority_fee_per_gas (optional, default is None)
        :type max_priority_fee_per_gas: Optional[int]
        :return: transaction hash
        :rtype: str
        """

        cancel = self.contract.functions.cancel(id)

        return self._default_transaction_handler(
            instantiated_contract_function=cancel,
            gas=gas,
            nonce=nonce,
            max_fee_per_gas=max_fee_per_gas,
            max_priority_fee_per_gas=max_priority_fee_per_gas
        )

    # batchOffer(payAmts (uint[]), payGems (address[]), buyAmts (uint[]), buyGems (address[])) -> uint256[]
    def batch_offer(
        self,
        pay_amts: List[int],
        pay_gems: List[ChecksumAddress],
        buy_amts: List[int],
        buy_gems: List[ChecksumAddress],
        nonce: Optional[int] = None,
        gas: int = 350000,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None
    ) -> str:
        """Batch the placement of a set of offers in one transaction

        :param pay_amts: the amounts of the token being sold
        :type pay_amts: List[int]
        :param pay_gems: the addresses of the tokens being sold
        :type pay_gems: List[ChecksumAddress]
        :param buy_amts: the amounts of the token being bought
        :type buy_amts: List[int]
        :param buy_gems: the addresses of the tokens being bought
        :type buy_gems: List[ChecksumAddress]
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce.
            (optional, default is None)
        :type nonce: Optional[int]
        :param gas: gas limit of the transaction, defaults to a very high estimate made when writing the class
        :type gas: int
        :param max_fee_per_gas: max fee that can be paid for gas, defaults to
            max_priority_fee (from chain) + (2 * base fee per gas of latest block) (optional, default is None)
        :type max_fee_per_gas: Optional[int]
        :param max_priority_fee_per_gas: max priority fee that can be paid for gas, defaults to calling the chain to
            estimate the max_priority_fee_per_gas (optional, default is None)
        :type max_priority_fee_per_gas: Optional[int]
        :return: transaction hash
        :rtype: str
        """
        if not (len(pay_amts) == len(pay_gems) == len(buy_amts) == len(buy_gems)):
            raise Exception("mismatches lengths in pay_amts, pay_gems, buy_amts and buy_gems")

        batch_offer = self.contract.functions.batchOffer(pay_amts, pay_gems, buy_amts, buy_gems)

        return self._default_transaction_handler(
            instantiated_contract_function=batch_offer,
            gas=gas,
            nonce=nonce,
            max_fee_per_gas=max_fee_per_gas,
            max_priority_fee_per_gas=max_priority_fee_per_gas
        )

    # batchCancel (ids (uint256[])) -> bool[]
    def batch_cancel(
        self,
        ids: List[int],
        nonce: Optional[int] = None,
        gas: int = 350000,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None
    ) -> str:
        """Cancel a set offer by offer id in a single transaction

        :param ids: the ids of the offers to cancel
        :type ids: List[int]
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce.
            (optional, default is None)
        :type nonce: Optional[int]
        :param gas: gas limit of the transaction, defaults to a very high estimate made when writing the class
        :type gas: int
        :param max_fee_per_gas: max fee that can be paid for gas, defaults to
            max_priority_fee (from chain) + (2 * base fee per gas of latest block) (optional, default is None)
        :type max_fee_per_gas: Optional[int]
        :param max_priority_fee_per_gas: max priority fee that can be paid for gas, defaults to calling the chain to
            estimate the max_priority_fee_per_gas (optional, default is None)
        :type max_priority_fee_per_gas: Optional[int]
        :return: transaction hash
        :rtype: str
        """
        cancels = self.contract.functions.cancel(ids)

        return self._default_transaction_handler(
            instantiated_contract_function=cancels,
            gas=gas,
            nonce=nonce,
            max_fee_per_gas=max_fee_per_gas,
            max_priority_fee_per_gas=max_priority_fee_per_gas
        )

    # batchRequote (ids (uint256[]), payAmts (uint[]), payGems (address[]), buyAmts (uint[]), buyGems (address[]))
    # -> uint256[]
    def batch_requote(
        self,
        ids: List[int],
        pay_amts: List[int],
        pay_gems: List[ChecksumAddress],
        buy_amts: List[int],
        buy_gems: List[ChecksumAddress],
        nonce: Optional[int] = None,
        gas: int = 350000,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None
    ) -> str:
        """Batch update a set of offers in a single transaction and return a list of new offer ids

        :param ids: the ids of the offers to cancel
        :type ids: List[int]
        :param pay_amts: the amounts of the token being sold
        :type pay_amts: List[int]
        :param pay_gems: the addresses of the tokens being sold
        :type pay_gems: List[ChecksumAddress]
        :param buy_amts: the amounts of the token being bought
        :type buy_amts: List[int]
        :param buy_gems: the addresses of the tokens being bought
        :type buy_gems: List[ChecksumAddress]
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce.
            (optional, default is None)
        :type nonce: Optional[int]
        :param gas: gas limit of the transaction, defaults to a very high estimate made when writing the class
        :type gas: int
        :param max_fee_per_gas: max fee that can be paid for gas, defaults to
            max_priority_fee (from chain) + (2 * base fee per gas of latest block) (optional, default is None)
        :type max_fee_per_gas: Optional[int]
        :param max_priority_fee_per_gas: max priority fee that can be paid for gas, defaults to calling the chain to
            estimate the max_priority_fee_per_gas (optional, default is None)
        :type max_priority_fee_per_gas: Optional[int]
        :return: transaction hash
        :rtype: str
        """

        batch_requote = self.contract.functions.batchRequote(ids, pay_amts, pay_gems, buy_amts, buy_gems)

        return self._default_transaction_handler(
            instantiated_contract_function=batch_requote,
            gas=gas,
            nonce=nonce,
            max_fee_per_gas=max_fee_per_gas,
            max_priority_fee_per_gas=max_priority_fee_per_gas
        )

    # sellAllAmount(pay_gem (address), pay_amt (uint256), buy_gem (address), min_fill_amount (uint256)) -> uint256
    def sell_all_amount(
        self,
        pay_gem: ChecksumAddress,
        pay_amt: int,
        buy_gem: ChecksumAddress,
        min_fill_amount: int,
        nonce: Optional[int] = None,
        gas: int = 350000,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None
    ) -> str:
        """Sell the pay_amt of the pay_gem token in exchange for buy_gem, on the condition that you receive at least the
        min_fill_amount of the buy_gem token

        :param pay_gem: the address of the tokens being sold
        :type pay_gem: ChecksumAddress
        :param pay_amt: the amount of the token being sold
        :type pay_amt: int
        :param buy_gem: the address of the tokens being bought
        :type buy_gem: ChecksumAddress
        :param min_fill_amount: minimum amount of the buy_gem token you want to receive
        :type min_fill_amount: int
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce.
            (optional, default is None)
        :type nonce: Optional[int]
        :param gas: gas limit of the transaction, defaults to a very high estimate made when writing the class
        :type gas: int
        :param max_fee_per_gas: max fee that can be paid for gas, defaults to
            max_priority_fee (from chain) + (2 * base fee per gas of latest block) (optional, default is None)
        :type max_fee_per_gas: Optional[int]
        :param max_priority_fee_per_gas: max priority fee that can be paid for gas, defaults to calling the chain to
            estimate the max_priority_fee_per_gas (optional, default is None)
        :type max_priority_fee_per_gas: Optional[int]
        :return: transaction hash
        :rtype: str
        """
        sell_all_amount = self.contract.functions.sellAllAmount(pay_gem, pay_amt, buy_gem, min_fill_amount)

        return self._default_transaction_handler(
            instantiated_contract_function=sell_all_amount,
            gas=gas,
            nonce=nonce,
            max_fee_per_gas=max_fee_per_gas,
            max_priority_fee_per_gas=max_priority_fee_per_gas
        )

    # buyAllAmount(buy_gem (address), buy_amt (uint256), pay_gem (address), max_fill_amount (uint256)) -> uint256
    def buy_all_amount(
        self,
        buy_gem: ChecksumAddress,
        buy_amt: int,
        pay_gem: ChecksumAddress,
        max_fill_amount: int,
        nonce: Optional[int] = None,
        gas: int = 350000,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None
    ) -> str:
        """Buy the buy_amt of the buy_gem token in exchange for pay_gem, on the condition that it does not exceed the
        max_fill_amount of the pay_gem token

        :param buy_gem: the address of the tokens being bought
        :type buy_gem: ChecksumAddress
        :param buy_amt: the amount of the token being bought
        :type buy_amt: int
        :param pay_gem: the address of the tokens being sold
        :type pay_gem: ChecksumAddress
        :param max_fill_amount: maximum amount of the pay_gem token you want to pay
        :type max_fill_amount: int
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce.
            (optional, default is None)
        :type nonce: Optional[int]
        :param gas: gas limit of the transaction, defaults to a very high estimate made when writing the class
        :type gas: int
        :param max_fee_per_gas: max fee that can be paid for gas, defaults to
            max_priority_fee (from chain) + (2 * base fee per gas of latest block) (optional, default is None)
        :type max_fee_per_gas: Optional[int]
        :param max_priority_fee_per_gas: max priority fee that can be paid for gas, defaults to calling the chain to
            estimate the max_priority_fee_per_gas (optional, default is None)
        :type max_priority_fee_per_gas: Optional[int]
        :return: transaction hash
        :rtype: str
        """
        buy_all_amount = self.contract.functions.sellAllAmount(buy_gem, buy_amt, pay_gem, max_fill_amount)

        return self._default_transaction_handler(
            instantiated_contract_function=buy_all_amount,
            gas=gas,
            nonce=nonce,
            max_fee_per_gas=max_fee_per_gas,
            max_priority_fee_per_gas=max_priority_fee_per_gas
        )



from typing import Optional, Tuple, List

from eth_typing import ChecksumAddress
from web3 import Web3
from web3.contract import Contract

from rubi.contracts.base_contract import BaseContract
from rubi.network import Network


class RubiconRouter(BaseContract):
    """This class represents the RubiconRouter.sol contract and by default has read functionality.
    If a wallet and key are passed in instantiation then this class can also be used to write to the contract instance.

    :param w3: Web3 instance
    :type w3: Web3
    :param contract: Contract instance
    :type contract: Contract
    :param wallet: a wallet address of the signer (optional, default is None)
    :type wallet: Optional[ChecksumAddress]
    :param key: the private key of the signer (optional, default is None)
    :type key: Optional[str]
    """

    def __init__(
        self,
        w3: Web3,
        contract: Contract,
        wallet: Optional[ChecksumAddress] = None,
        key: Optional[str] = None
    ) -> None:
        """constructor method"""
        super().__init__(
            w3=w3,
            contract=contract,
            wallet=wallet,
            key=key
        )

    @classmethod
    def from_network(
        cls,
        network: Network,
        wallet: Optional[ChecksumAddress] = None,
        key: Optional[str] = None
    ) -> "RubiconRouter":
        """Create a RubiconRouter instance based on a Network instance.

        :param network: A Network instance.
        :type network: Network
        :param wallet: Optional wallet address to use for interacting with the contract (optional, default is None). 
        :type wallet: Optional[ChecksumAddress]
        :param key: Optional private key for the wallet (optional, default is None).
        :type key: Optional[str]
        :return: A RubiconRouter instance based on the Network instance.
        :rtype: RubiconRouter
        """
        return cls.from_address_and_abi(
            w3=network.w3,
            address=network.rubicon.router.address,
            contract_abi=network.rubicon.router.abi,
            wallet=wallet,
            key=key
        )

    ######################################################################
    # read calls
    ######################################################################

    # getMakerBalance(baseToken (address), tokens (List[address]), maker (address))
    # -> (uint256 balanceInBook, uint256 balance)
    def get_maker_balance(
        self,
        base_token: ChecksumAddress,
        tokens: List[ChecksumAddress],
        maker: ChecksumAddress,
    ) -> Tuple[int, int]:
        """Iterates through all the base_token/tokens[i] offers of the maker and returns the balance of the base_token
        in the book and the balance of the base token.

        :param base_token: The address of the base token.
        :type base_token: ChecksumAddress
        :param tokens: A list of all the tokens to calculate the balance of
        :type tokens: List[ChecksumAddress]
        :param maker: The address of the maker to fet the balance for
        :type maker: ChecksumAddress
        :return: balance in book, total token balance
        :rtype: Tuple[int, int]
        """

        return self.contract.functions.getMakerBalance(base_token, tokens, maker).call()

    # getMakerBalanceInPair(asset (address), quote (address), maker (address)) -> (uint256 balance)
    def get_maker_balance_in_pair(
        self,
        asset: ChecksumAddress,
        quote: ChecksumAddress,
        maker: ChecksumAddress,
    ) -> int:
        """Retrieves the balance of a specific maker for a given asset/quote pair.

        :param asset: The address of the asset token.
        :type asset: ChecksumAddress
        :param quote: The address of the quote token.
        :type quote: ChecksumAddress
        :param maker: The address of the maker.
        :type maker: ChecksumAddress
        :return: The balance of the maker in the specified asset/quote pair.
        :rtype: int
        """

        return self.contract.functions.getMakerBalanceInPair(asset, quote, maker).call()

    # getBookFromPair(asset (address), quote (address)) -> (uint256[3][] asks, uint256[3][] bids)
    def get_book_from_pair(
        self,
        asset: ChecksumAddress,
        quote: ChecksumAddress,
    ) -> Tuple[List[List[int]], List[List[int]]]:
        """Retrieves the order book for a specific asset/quote pair.

        :param asset: The address of the asset token.
        :type asset: ChecksumAddress
        :param quote: The address of the quote token.
        :type quote: ChecksumAddress
        :return: A tuple containing two lists: asks and bids. Each list contains a sublist of length 3, representing
            the order book entries in the following format (pay_amt, buy_amt, id). The asks list represents the orders
            selling the asset, while the bids list represents the orders buying the asset.
        :rtype: Tuple[List[List[int]], List[List[int]]]
        """

        return self.contract.functions.getBookFromPair(asset, quote).call()

    # getBookDepth(tokenIn (address), tokenOut (address)) -> (uint256 depth, uint256 bestOfferID)
    def get_book_depth(self, token_in: ChecksumAddress, token_out: ChecksumAddress) -> Tuple[int, int]:
        """Retrieves the depth of one side of the order book for a specific token pair along with the id of the best
        offer.

        :param token_in: The address of the quote.
        :type token_in: ChecksumAddress
        :param token_out: The address of the asset.
        :type token_out: ChecksumAddress
        :return: A tuple containing the depth of the order book and the ID of the best offer for token_out/token_in.
        :rtype: Tuple[int, int]
        """

        return self.contract.functions.getBookDepth(token_in, token_out).call()

    # getBestOfferAndInfo(asset (address), quote (address)) -> (uint256 id, uint256, address, uint256, address)
    def get_best_offer_and_info(
        self,
        asset: ChecksumAddress,
        quote: ChecksumAddress
    ) -> Tuple[int, int, ChecksumAddress, int, ChecksumAddress]:
        """Retrieves the information and id of the best offer for a specific asset/quote pair.

        :param asset: The address of the asset token.
        :type asset: ChecksumAddress
        :param quote: The address of the quote token.
        :type quote: ChecksumAddress
        :return: A tuple containing the ID of the best offer, the pay_amt, the address of the pay_gem,
            the buy_amt, and the address of the buy_gem.
        :rtype: Tuple[int, int, ChecksumAddress, int, ChecksumAddress]
        """

        return self.contract.functions.getBestOfferAndInfo(asset, quote).call()

    # getExpectedSwapFill(pay_amt (uint256), buy_amt_min (uint256), route (address[])) -> (uint256 amount)
    def get_expected_swap_fill(self, pay_amt: int, buy_amt_min: int, route: List[ChecksumAddress]) -> int:
        """Estimates the expected amount including fees when swapping the specified payment amount using the specified
        route. reverts with an exception if the swap cannot achieve the buy_amt_min

        :param pay_amt: The payment amount.
        :type pay_amt: int
        :param buy_amt_min: The minimum buy amount.
        :type buy_amt_min: int
        :param route: The route of addresses representing the swap path.
        :type route: List[ChecksumAddress]
        :return: The estimated swap amount including fees.
        :rtype: int
        """

        return self.contract.functions.getExpectedSwapFill(pay_amt, buy_amt_min, route).call()

    # getExpectedMultiswapFill(pay_amts (uint256[]), buy_amt_mins (uint256[]), routes (address[][]))
    # -> (uint256 amount)
    def get_expected_multiswap_fill(
        self,
        pay_amts: List[int],
        buy_amt_mins: List[int],
        routes: List[List[ChecksumAddress]]
    ) -> int:
        """Estimates the expected amount including fees when swapping multiple specified payment amount using multiple
        specified routes. reverts with an exception if the multiswap cannot achieve the buy_amt_mins along each route

        :param pay_amts: The list of payment amounts for each swap.
        :type pay_amts: List[int]
        :param buy_amt_mins: The list of minimum buy amounts for each swap.
        :type buy_amt_mins: List[int]
        :param routes: The list of routes, where each route is a list of addresses representing the swap path.
        :type routes: List[List[ChecksumAddress]]
        :return: The estimated multi-swap amount.
        :rtype: int
        """

        return self.contract.functions.getExpectedMultiswapFill(pay_amts, buy_amt_mins, routes).call()

    # checkClaimAllUserBonusTokens(address user, address[] targetBathTokens, address token)
    # -> (uint256 earnedAcrossPools)
    def check_claim_all_user_bonus_tokens(
        self,
        user: ChecksumAddress,
        target_bath_tokens: List[ChecksumAddress],
        token: ChecksumAddress
    ) -> int:
        """Checks the all bonus tokens that can be claimed by a user earned across all specified rubicon pools.

        :param user: The address of the user.
        :type user: ChecksumAddress
        :param target_bath_tokens: The list of target bath tokens to claim bonus from.
        :type target_bath_tokens: List[ChecksumAddress]
        :param token: The address of the token for which the bonus is claimed.
        :type token: ChecksumAddress
        :return: The total amount earned across all pools.
        :rtype: int
        """

        return self.contract.functions.checkClaimAllUserBonusTokens(user, target_bath_tokens, token).call()

    ######################################################################
    # write calls
    ######################################################################

    # multiswap(address[][] routes, uint256[] pay_amts, uint256[] buy_amts_min, address to) -> bool
    def multiswap(
        self,
        routes: List[List[ChecksumAddress]],
        pay_amts: List[int],
        buy_amts_min: List[int],
        to: ChecksumAddress,
        nonce: Optional[int] = None,
        gas: int = 350000,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None
    ) -> str:
        """Perform a multiple swaps for the specified payment amounts using the specified routes. Reverts with an
        exception if any of the swaps cannot achieve the buy_amt_min along the specified route.

        :param routes: The list of routes, where each route is a list of addresses representing the swap path.
        :type routes: List[List[ChecksumAddress]]
        :param pay_amts: The list of payment amounts for each swap.
        :type pay_amts: List[int]
        :param buy_amts_min: The list of minimum buy amounts for each swap.
        :type buy_amts_min: List[int]
        :param to: The address of the recipient.
        :type to: ChecksumAddress
        :param nonce: Nonce of the transaction. Defaults to calling the chain state to get the nonce.
            (optional, default is None).
        :type nonce: Optional[int]
        :param gas: Gas limit of the transaction. Defaults to a very high estimate made when writing the class.
        :type gas: int
        :param max_fee_per_gas: Max fee that can be paid for gas. Defaults to max_priority_fee (from chain)
            + (2 * base fee per gas of latest block) (optional, default is None).
        :type max_fee_per_gas: Optional[int]
        :param max_priority_fee_per_gas: Max priority fee that can be paid for gas. Defaults to calling the chain to
            estimate the max_priority_fee_per_gas (optional, default is None).
        :type max_priority_fee_per_gas: Optional[int]
        :return: Transaction hash.
        :rtype: str
        """

        multiswap = self.contract.functions.multiswap(routes, pay_amts, buy_amts_min, to)

        return self._default_transaction_handler(
            instantiated_contract_function=multiswap,
            gas=gas,
            nonce=nonce,
            max_fee_per_gas=max_fee_per_gas,
            max_priority_fee_per_gas=max_priority_fee_per_gas
        )

    # swap(uint256 pay_amt, uint256 buy_amt_min, address[] memory route, address to) -> uint256
    def swap(
        self,
        pay_amt: int,
        buy_amt_min: int,
        route: List[ChecksumAddress],
        to: ChecksumAddress,
        nonce: Optional[int] = None,
        gas: int = 350000,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None
    ) -> str:
        """Perform a swap operation with the specified payment amount using the specified route and paying out to the
        recipient. Reverts if the swap does not result in the buy_min_amount.

        :param pay_amt: The payment amount.
        :type pay_amt: int
        :param buy_amt_min: The minimum buy amount.
        :type buy_amt_min: int
        :param route: The route, represented as a list of addresses representing the swap path.
        :type route: List[ChecksumAddress]
        :param to: The address of the recipient.
        :type to: ChecksumAddress
        :param nonce: Nonce of the transaction. Defaults to calling the chain state to get the nonce.
            (optional, default is None).
        :type nonce: Optional[int]
        :param gas: Gas limit of the transaction. Defaults to a very high estimate made when writing the class.
        :type gas: int
        :param max_fee_per_gas: Max fee that can be paid for gas. Defaults to max_priority_fee (from chain)
            + (2 * base fee per gas of latest block) (optional, default is None).
        :type max_fee_per_gas: Optional[int]
        :param max_priority_fee_per_gas: Max priority fee that can be paid for gas. Defaults to calling the chain to
            estimate the max_priority_fee_per_gas (optional, default is None).
        :type max_priority_fee_per_gas: Optional[int]
        :return: Transaction hash.
        :rtype: str
        """

        swap = self.contract.functions.swap(pay_amt, buy_amt_min, route, to)

        return self._default_transaction_handler(
            instantiated_contract_function=swap,
            gas=gas,
            nonce=nonce,
            max_fee_per_gas=max_fee_per_gas,
            max_priority_fee_per_gas=max_priority_fee_per_gas
        )

    # TODO
    #  sellAllAmountForETH
    #  sellAllAmountWithETH
    #  buyAllAmountWithETH
    #  buyAllAmountForETH
    #  swapWithETH
    #  swapForETH
    #  offerWithETH
    #  offerForETH
    #  cancelForETH
    #  depositWithETH
    #  withdrawForETH



import os
import json
import hexbytes
import logging as log
from eth_abi import decode
from eth_abi.codec import ABICodec
from web3._utils.events import get_event_data

from rubi.contracts.helper import networks


# DEPRECATED
# TODO: this is the old implementation for the RubiconMarketAidFactory and RubiconMarketAid. It needs to be updated in
#  line with all the other contracts in this repo. Usage is not recommended in it's current state and it is not included
#  in any of the __init__.py files.
class FactoryAid:
    """this class represents the MarketAidFactory.sol contract and has read functionality for the contract
    
    :param w3: Web3 instance
    :type w3: Web3
    :param contract: an optional contract instance, if not provided, the contract will be instantiated using the address and abi from the networks.py file given the chain id of the w3 instance
    :type contract: Web3 object, optional
    """

    def __init__(self, w3, contract=None):
        """constructor method"""
        raise DeprecationWarning()

        chain = w3.eth.chain_id

        if contract:
            self.contract = contract
            self.address = self.contract.address
        else:
            # TODO: add error handling for unsupported chains
            network = networks[chain]()
            self.contract = w3.eth.contract(address=network.factory, abi=network.factory_abi)
            self.address = network.factory

        # set the class variables
        self.chain = chain
        self.w3 = w3

    ######################################################################
    # read calls
    ######################################################################

    # admin()
    def admin(self):
        """returns the admin address of the contract

        :return: the admin address
        :rtype: str
        """

        try:
            admin = self.contract.functions.admin().call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return admin

    # getUserMarketAids(user (address))
    def get_user_market_aids(self, user):
        """returns the market aid addresses for a given user

        :return: an array of market aid addresses
        :rtype: list
        """

        try:
            aids = self.contract.functions.getUserMarketAids(user).call()
        except ValueError:
            aids = self.contract.functions.getUserMarketAids(self.w3.to_checksum_address(user)).call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return aids

    # rubiconMarket()
    def rubicon_market(self):
        """returns the address of the RubiconMarket contract

        :return: rubicon market address
        :rtype: str
        """

        try:
            rubicon_market = self.contract.functions.rubiconMarket().call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return rubicon_market

class FactoryAidSigner(FactoryAid):
    """this class represents the MarketAidFactory.sol contract and has read and write functionality for the contract. this class inherits from the FactoryAid class
    
    :param w3: a web3 instance
    :type w3: Web3
    :param wallet: the signers wallet address
    :type wallet: str
    :param key: the signers private key
    :type key: str
    :param contract: an optional parameter that allows you to pass in a contract instance, if none the contract will be instantiated from the rolodex.py file
    :type contract_address: str, optional  
    """

    def __init__(self, w3, wallet, key, contract=None):
        super().__init__(w3, contract)
        self.wallet = wallet
        self.key = key

    ######################################################################
    # write calls
    ######################################################################

    # createMarketAidInstance()
    def create_market_aid_instance(self, nonce=None, gas=3000000, gas_price=None):
        """this function creates a new market aid instance

        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce
        :type nonce: int, optional
        :param gas: gas limit of the transaction, defaults to a value of 3000000
        :type gas: int, optional
        :param gas_price: gas price of the transaction, defaults to the gas price of the chain
        :type gas_price: int, optional
        :return: the transaction object of the create market aid instance transaction, None if there is an error
        :rtype: dict, None
        """

        if nonce is None:
            txn_nonce = self.w3.eth.get_transaction_count(self.wallet)
        else:
            txn_nonce = nonce

        if gas_price is None:
            gas_price = self.w3.eth.gas_price

        txn = {'chainId': self.chain, 'gas' : gas, 'gasPrice': gas_price, 'nonce': txn_nonce}

        try:
            create = self.contract.functions.createMarketAidInstance().build_transaction(txn)
            create = self.w3.eth.account.sign_transaction(create, self.key)
            self.w3.eth.send_raw_transaction(create.rawTransaction)

            # if a user is not providing a nonce, wait for the transaction to either be confirmed or rejected before continuing
            if nonce is None:
                if self.w3.eth.wait_for_transaction_receipt(create.hash)['status'] == 0:
                    log.error(f"create_market_aid_instance transaction failed: {create.hash.hex()}")
                    raise SystemExit()

        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return create

class MarketAid:

    # init function
    # TODO: possibly allow a chain variable to be passed in
    def __init__(self, w3, address, contract=None):
        raise DeprecationWarning()

        # load in the aid abi 
        path = f"{os.path.dirname(os.path.realpath(__file__))}/helper/abis/MarketAid.json"
        with open(path) as f:
            aid_abi = json.load(f)
        f.close()

        # create contract instance or set based upon initiliazation
        if contract:
            self.contract = contract
            self.address = self.contract.address
            chain = w3.eth.chain_id
        else:
            chain = w3.eth.chain_id
            #network = networks[chain]()
            self.contract = w3.eth.contract(address=address, abi=aid_abi)
            self.address = address

        # set the class variables
        self.chain = chain
        self.w3 = w3
        self.log_strategist_trade_abi = self.contract.events.LogStrategistTrade()._get_event_abi()
        self.log_scrubbed_strat_trade_abi = self.contract.events.LogScrubbedStratTrade()._get_event_abi()
        self.log_batch_market_making_trades_abi = self.contract.events.LogBatchMarketMakingTrades()._get_event_abi()
        self.log_requote_abi = self.contract.events.LogRequote()._get_event_abi()
        self.log_batch_requote_offers_abi = self.contract.events.LogBatchRequoteOffers()._get_event_abi()
        self.codec: ABICodec = w3.codec

    ######################################################################
    # read calls
    ######################################################################

    # RubiconMarketAddress()
    def rubicon_market_address(self):
        """this function returns the address of the RubiconMarket contract

        :return: the address of the RubiconMarket contract
        :rtype: str
        """

        try:
            rubicon_market_address = self.contract.functions.RubiconMarketAddress().call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return rubicon_market_address

    # admin()
    def admin(self):
        """this function returns the address of the admin

        :return: the address of the admin
        :rtype: str
        """

        try:
            admin = self.contract.functions.admin().call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return admin

    # approvedStrategists()
    def approved_strategists(self, address):
        """this function returns the list of approved strategists

        :return: the list of approved strategists
        :rtype: list
        """

        try:
            approved_strategists = self.contract.functions.approvedStrategists(address).call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return approved_strategists

    # getOutstandingStrategistTrades(asset (address), quote (address), strategist (address))
    def get_outstanding_strategist_trades(self, asset, quote, strategist):
        """this function returns the list of outstanding trades for a strategist

        :param asset: the address of the asset
        :type asset: str
        :param quote: the address of the quote
        :type quote: str
        :param strategist: the address of the strategist
        :type strategist: str
        :return: the list of outstanding trades for a strategist
        :rtype: list
        """

        try:
            outstanding_strategist_trades = self.contract.functions.getOutstandingStrategistTrades(asset, quote, strategist).call()
        except ValueError:
            outstanding_strategist_trades = self.contract.functions.getOutstandingStrategistTrades(self.w3.to_checksum_address(asset), self.w3.to_checksum_address(quote), self.w3.to_checksum_address(strategist)).call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return outstanding_strategist_trades

    # getStrategistTotalLiquidity(asset (address), quote (address), strategist (address))
    def get_strategist_total_liquidity(self, asset, quote, strategist):
        """this function returns the total liquidity for a strategist 

        :param asset: the address of the asset
        :type asset: str
        :param quote: the address of the quote
        :type quote: str
        :param strategist: the address of the strategist
        :type strategist: str
        :return: the total liquidity for a strategist
        :rtype: int
        """

        try:
            strategist_total_liquidity = self.contract.functions.getStrategistTotalLiquidity(asset, quote, strategist).call()
        except ValueError:
            strategist_total_liquidity = self.contract.functions.getStrategistTotalLiquidity(self.w3.to_checksum_address(asset), self.w3.to_checksum_address(quote), self.w3.to_checksum_address(strategist)).call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return strategist_total_liquidity

    # isApprovedStrategist(strategist (address))
    def is_approved_strategist(self, strategist):
        """this function returns whether or not a strategist is approved

        :param strategist: the address of the strategist
        :type strategist: str
        :return: whether or not a strategist is approved
        :rtype: bool
        """

        try:
            is_approved_strategist = self.contract.functions.isApprovedStrategist(strategist).call()
        except ValueError:
            is_approved_strategist = self.contract.functions.isApprovedStrategist(self.w3.to_checksum_address(strategist)).call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return is_approved_strategist

    def get_strategist_trade(self, trade_id):
        """this function returns a strategist trade

        :param trade_id: the id of the trade
        :type trade_id: int
        :return: a an array containing the relevant information for a pair of offers: [ask_id, ask_pay_amt, ask_asset, bid_id, bid_pay_amt, bid_asset, strategist, timestamp]
        :rtype: dict
        """

        # check that the trade id is an integer
        if not isinstance(trade_id, int):
            log.error("trade id is not an integer")
            return None

        try:
            strategist_trade = self.contract.functions.strategistTrades(trade_id).call()
        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return strategist_trade

    ######################################################################
    # events & helpers
    ######################################################################

    # TODO: today the event signature is hardcoded, but we should be able to get it from the contract
    # the graph does something similar to this when you run codegen, it should be a simple string manipulation problem from the abis 
    def get_log_strategist_trade_hash(self):
        return self.w3.keccak(text="LogStrategistTrade(uint256,bytes32,bytes32,address,address,uint256,address)").hex()

    # TODO: determine if this is the right assumtption to make about how the data is being received 
    # this assumes that the function is being used directly in the context of being passed raw data from a websocket stream that has been loaded and converted to an AttributeDict
    # TODO: i feel like this could be done much faster... tracing will tell us
    def stream_log_strategist_trade(self, data):

        # load the data into an attribute dictionary that web3 can use
        # data = AttributeDict(json.loads(data))

        # convert the topics, transaction hash, and block hash to hex strings
        data['params']['result']['topics'] = [hexbytes.HexBytes(topic) for topic in data['params']['result']['topics']]
        data['params']['result']['transactionHash'] = hexbytes.HexBytes(data['params']['result']['transactionHash'])
        data['params']['result']['blockHash'] = hexbytes.HexBytes(data['params']['result']['blockHash'])

        # get the event data from the log
        try:
            event = get_event_data(self.codec, self.log_strategist_trade_abi, data['params']['result'])

            # decode the offer id
            # TODO: there is probably a way to do this that does not hardcode the type of the id
            trade_id = decode(['uint256'], event['args']['strategistTradeID'])[0]

            # now pass an offer back in the form of a dictionary
            # TODO: this is probably not the most performant, we will optimize later
            trade = {
                    'id': trade_id,
                    'txn': event['transactionHash'].hex(),
                    'event': event['event'],
                    'ask_id': event['args']['askId'],
                    'bid_id': event['args']['bidId'],
                    'ask_asset': event['args']['askAsset'],
                    'bid_asset': event['args']['bidAsset'],
                    'timestamp': event['args']['timestamp'],
                    'owner': event['args']['strategist']
            }
            return trade

        except Exception as e:
            log.error(e, exc_info=True)
            return None

    def parse_log_strategist_trade(self, log):

        # get the event data from the log
        try:
            event = get_event_data(self.codec, self.log_strategist_trade_abi, log['params']['result'])

            # decode the offer id
            # TODO: there is probably a way to do this that does not hardcode the type of the id
            trade_id = decode(['uint256'], event['args']['strategistTradeID'])[0]

            # now pass an offer back in the form of a dictionary
            # TODO: this is probably not the most performant, we will optimize later
            trade = {
                    'id': trade_id,
                    'txn': event['transactionHash'].hex(),
                    'event': event['event'],
                    'ask_id': event['args']['askId'],
                    'bid_id': event['args']['bidId'],
                    'ask_asset': event['args']['askAsset'],
                    'bid_asset': event['args']['bidAsset'],
                    'timestamp': event['args']['timestamp'],
                    'owner': event['args']['strategist']
            }
            return trade

        except Exception as e:
            log.error(e, exc_info=True)
            return None

    '''
    event LogScrubbedStratTrade(
        uint256 strategistIDScrubbed,
        uint256 assetFill,
        address bathAssetAddress,
        uint256 quoteFill,
        address quoteAddress
    );
    '''

    # TODO: today the event signature is hardcoded, but we should be able to get it from the contract
    # the graph does something similar to this when you run codegen, it should be a simple string manipulation problem from the abis 
    def get_log_scrubbed_strat_trade_hash(self):
        return self.w3.keccak(text="LogStrategistTrade(uint256,uint256,address,uint256,address)").hex()

    # TODO: determine if this is the right assumtption to make about how the data is being received 
    # this assumes that the function is being used directly in the context of being passed raw data from a websocket stream that has been loaded and converted to an AttributeDict
    # TODO: i feel like this could be done much faster... tracing will tell us
    def stream_log_scrubbed_strat_trade(self, data):

        # load the data into an attribute dictionary that web3 can use
        # data = AttributeDict(json.loads(data))

        # convert the topics, transaction hash, and block hash to hex strings
        data['params']['result']['topics'] = [hexbytes.HexBytes(topic) for topic in data['params']['result']['topics']]
        data['params']['result']['transactionHash'] = hexbytes.HexBytes(data['params']['result']['transactionHash'])
        data['params']['result']['blockHash'] = hexbytes.HexBytes(data['params']['result']['blockHash'])

        # get the event data from the log
        try:
            event = get_event_data(self.codec, self.log_scrubbed_strat_trade_abi, data['params']['result'])

            # decode the offer id
            # TODO: there is probably a way to do this that does not hardcode the type of the id
            trade_id = decode(['uint256'], event['args']['strategistIDScrubbed'])[0]

            # now pass an offer back in the form of a dictionary
            # TODO: this is probably not the most performant, we will optimize later
            trade = {
                    'id': trade_id,
                    'txn': event['transactionHash'].hex(),
                    'event': event['event'],
                    'asset_fill': event['args']['assetFill'],
                    'bath_asset_address': event['args']['bathAssetAddress'],
                    'quote_fill': event['args']['quoteFill'],
                    'quote_address': event['args']['quoteAddress']
            }
            return trade

        except Exception as e:
            log.error(e, exc_info=True)
            return None

    def parse_log_scrubbed_strat_trade(self, log):

        # get the event data from the log
        try:
            event = get_event_data(self.codec, self.log_scrubbed_strat_trade_abi, log['params']['result'])

            # decode the offer id
            # TODO: there is probably a way to do this that does not hardcode the type of the id
            trade_id = decode(['uint256'], event['args']['strategistIDScrubbed'])[0]

            # now pass an offer back in the form of a dictionary
            # TODO: this is probably not the most performant, we will optimize later
            trade = {
                    'id': trade_id,
                    'txn': event['transactionHash'].hex(),
                    'event': event['event'],
                    'asset_fill': event['args']['assetFill'],
                    'bath_asset_address': event['args']['bathAssetAddress'],
                    'quote_fill': event['args']['quoteFill'],
                    'quote_address': event['args']['quoteAddress']
            }
            return trade

        except Exception as e:
            log.error(e, exc_info=True)
            return None

    '''
    event LogStrategistRewardClaim(
        address strategist,
        address asset,
        uint256 amountOfReward,
        uint256 timestamp
    );
    '''

    '''
    event LogBatchMarketMakingTrades(
        address strategist, 
        uint256[] trades
    );
    '''
    # TODO: today the event signature is hardcoded, but we should be able to get it from the contract
    # the graph does something similar to this when you run codegen, it should be a simple string manipulation problem from the abis 
    def get_log_batch_market_making_trades_hash(self):
        return self.w3.keccak(text="LogBatchMarketMakingTrades(address,uint256[])").hex()

    # TODO: determine if this is the right assumtption to make about how the data is being received 
    # this assumes that the function is being used directly in the context of being passed raw data from a websocket stream that has been loaded and converted to an AttributeDict
    # TODO: i feel like this could be done much faster... tracing will tell us
    def stream_log_batch_market_making_trades(self, data):

        # load the data into an attribute dictionary that web3 can use
        # data = AttributeDict(json.loads(data))

        # convert the topics, transaction hash, and block hash to hex strings
        data['params']['result']['topics'] = [hexbytes.HexBytes(topic) for topic in data['params']['result']['topics']]
        data['params']['result']['transactionHash'] = hexbytes.HexBytes(data['params']['result']['transactionHash'])
        data['params']['result']['blockHash'] = hexbytes.HexBytes(data['params']['result']['blockHash'])

        # get the event data from the log
        try:
            event = get_event_data(self.codec, self.log_batch_market_making_trades_abi, data['params']['result'])

            # now pass an offer back in the form of a dictionary
            # TODO: this is probably not the most performant, we will optimize later
            trade = {
                    'txn': event['transactionHash'].hex(),
                    'event': event['event'],
                    'strategist': event['args']['strategist'],
                    'trades': event['args']['trades']
            }
            return trade

        except Exception as e:
            log.error(e, exc_info=True)
            return None

    def parse_log_batch_market_making_trades(self, log):

        # get the event data from the log
        try:
            event = get_event_data(self.codec, self.log_batch_market_making_trades_abi, log['params']['result'])

            # now pass an offer back in the form of a dictionary
            # TODO: this is probably not the most performant, we will optimize later
            trade = {
                    'txn': event['transactionHash'].hex(),
                    'event': event['event'],
                    'strategist': event['args']['strategist'],
                    'trades': event['args']['trades']
            }
            return trade

        except Exception as e:
            log.error(e, exc_info=True)
            return None

    '''
        event LogRequote(
        address strategist,
        uint256 scrubbedOfferID,
        uint256 newOfferID
    );
    '''
    # TODO: today the event signature is hardcoded, but we should be able to get it from the contract
    # the graph does something similar to this when you run codegen, it should be a simple string manipulation problem from the abis 
    def get_log_requote_hash(self):
        return self.w3.keccak(text="LogRequote(address,uint256,uint256)").hex()

    # TODO: determine if this is the right assumtption to make about how the data is being received 
    # this assumes that the function is being used directly in the context of being passed raw data from a websocket stream that has been loaded and converted to an AttributeDict
    # TODO: i feel like this could be done much faster... tracing will tell us
    def stream_log_requote(self, data):

        # load the data into an attribute dictionary that web3 can use
        # data = AttributeDict(json.loads(data))

        # convert the topics, transaction hash, and block hash to hex strings
        data['params']['result']['topics'] = [hexbytes.HexBytes(topic) for topic in data['params']['result']['topics']]
        data['params']['result']['transactionHash'] = hexbytes.HexBytes(data['params']['result']['transactionHash'])
        data['params']['result']['blockHash'] = hexbytes.HexBytes(data['params']['result']['blockHash'])

        # get the event data from the log
        try:
            event = get_event_data(self.codec, self.log_requote_abi, data['params']['result'])

            # decode the offer id
            # TODO: there is probably a way to do this that does not hardcode the type of the id
            scrub_trade_id = decode(['uint256'], event['args']['scrubbedOfferID'])[0]
            trade_id = decode(['uint256'], event['args']['newOfferID'])[0]

            # now pass an offer back in the form of a dictionary
            # TODO: this is probably not the most performant, we will optimize later
            trade = {
                    'id': trade_id,
                    'txn': event['transactionHash'].hex(),
                    'event': event['event'],
                    'strategist': event['args']['strategist'],
                    'scrub_trade_id': scrub_trade_id,
                    'trade_id': trade_id
            }
            return trade

        except Exception as e:
            log.error(e, exc_info=True)
            return None

    def parse_log_requote(self, log):

        # get the event data from the log
        try:
            event = get_event_data(self.codec, self.log_requote_abi, log['params']['result'])

            # decode the offer id
            # TODO: there is probably a way to do this that does not hardcode the type of the id
            scrub_trade_id = decode(['uint256'], event['args']['scrubbedOfferID'])[0]
            trade_id = decode(['uint256'], event['args']['newOfferID'])[0]

            # now pass an offer back in the form of a dictionary
            # TODO: this is probably not the most performant, we will optimize later
            trade = {
                    'id': trade_id,
                    'txn': event['transactionHash'].hex(),
                    'event': event['event'],
                    'strategist': event['args']['strategist'],
                    'scrub_trade_id': scrub_trade_id,
                    'trade_id': trade_id
            }
            return trade

        except Exception as e:
            log.error(e, exc_info=True)
            return None

    '''
    event LogBatchRequoteOffers(address strategist, uint256[] scrubbedOfferIDs);
    '''
    # TODO: today the event signature is hardcoded, but we should be able to get it from the contract
    # the graph does something similar to this when you run codegen, it should be a simple string manipulation problem from the abis 
    def get_log_batch_requote_offers_hash(self):
        return self.w3.keccak(text="LogBatchRequoteOffers(address,uint256[])").hex()

    # TODO: determine if this is the right assumtption to make about how the data is being received 
    # this assumes that the function is being used directly in the context of being passed raw data from a websocket stream that has been loaded and converted to an AttributeDict
    # TODO: i feel like this could be done much faster... tracing will tell us
    def stream_log_batch_requote_offers(self, data):

        # load the data into an attribute dictionary that web3 can use
        # data = AttributeDict(json.loads(data))

        # convert the topics, transaction hash, and block hash to hex strings
        data['params']['result']['topics'] = [hexbytes.HexBytes(topic) for topic in data['params']['result']['topics']]
        data['params']['result']['transactionHash'] = hexbytes.HexBytes(data['params']['result']['transactionHash'])
        data['params']['result']['blockHash'] = hexbytes.HexBytes(data['params']['result']['blockHash'])

        # get the event data from the log
        try:
            event = get_event_data(self.codec, self.log_batch_requote_offers_abi, data['params']['result'])

            # now pass an offer back in the form of a dictionary
            # TODO: this is probably not the most performant, we will optimize later
            trade = {
                    'txn': event['transactionHash'].hex(),
                    'event': event['event'],
                    'strategist': event['args']['strategist'],
                    'scrub_trade_id': event['args']['scrubbedOfferIDs']
            }
            return trade

        except Exception as e:
            log.error(e, exc_info=True)
            return None

    def parse_log_batch_requote_offers(self, log):

        # get the event data from the log
        try:
            event = get_event_data(self.codec, self.log_batch_requote_offers_abi, log['params']['result'])

            # now pass an offer back in the form of a dictionary
            # TODO: this is probably not the most performant, we will optimize later
            trade = {
                    'txn': event['transactionHash'].hex(),
                    'event': event['event'],
                    'strategist': event['args']['strategist'],
                    'scrub_trade_id': event['args']['scrubbedOfferIDs']
            }
            return trade

        except Exception as e:
            log.error(e, exc_info=True)
            return None


class MarketAidSigner(MarketAid):
    """this class represents a MarketAid.sol contract with read and write functionality. it inherits from the MarketAid class and adds the ability to sign transactions.

    :param w3: a web3 instance
    :type w3: Web3
    :param address: the address of the contract
    :type address: str
    :param wallet: the signers wallet address
    :type wallet: str
    :param key: the signers private key
    :type key: str
    :param contract: an optional parameter that allows you to pass in a contract instance, if none the contract will be instantiated from the rolodex.py file
    :type contract_address: str, optional
    """

    def __init__(self, w3, address, wallet, key, contract=None):
        super().__init__(w3, address, contract)
        self.wallet = wallet
        self.key = key

    ######################################################################
    # write calls
    ######################################################################

    # adminMaxApproveTarget(target (address), token (address))
    def admin_max_approve_target(self, target, token, nonce=None, gas=3000000, gas_price=None):
        """this function sets the max approval for a target address to spend a token on behalf of the contract
        
        :param target: the address of the target
        :type target: str
        :param token: the address of the token
        :type token: str
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce
        :type nonce: int, optional
        :param gas: gas limit of the transaction, defaults to a value of 3000000
        :type gas: int, optional
        :param gas_price: gas price of the transaction, defaults to the gas price of the chain
        :type gas_price: int, optional
        :return: the transaction object of the approval transaction, returns None if the transaction fails
        :rtype: dict
        """

        if nonce is None:
            txn_nonce = self.w3.eth.get_transaction_count(self.wallet)
        else:
            txn_nonce = nonce

        if gas_price is None:
            gas_price = self.w3.eth.gas_price

        txn = {'chainId': self.chain, 'gas' : gas, 'gasPrice': gas_price, 'nonce': txn_nonce}

        try:
            max_approve = self.contract.functions.adminMaxApproveTarget(target, token).build_transaction(txn)
            max_approve = self.w3.eth.account.sign_transaction(max_approve, self.key)
            self.w3.eth.send_raw_transaction(max_approve.rawTransaction)

            # if a user is not providing a nonce, wait for the transaction to either be confirmed or rejected before continuing
            if nonce is None:
                if self.w3.eth.wait_for_transaction_receipt(max_approve.hash)['status'] == 0:
                    log.error(f'admin_max_approve_target transaction {max_approve.hash.hex()} failed')
                    raise SystemExit()

        except ValueError:
            log.warning('most likely a checksum error... retrying with checksummed addresses')
            max_approve = self.contract.functions.adminMaxApproveTarget(self.w3.to_checksum_address(target), self.w3.to_checksum_address(token)).build_transaction(txn)
            max_approve = self.w3.eth.account.sign_transaction(max_approve, self.key)
            self.w3.eth.send_raw_transaction(max_approve.rawTransaction)

            # if a user is not providing a nonce, wait for the transaction to either be confirmed or rejected before continuing
            if nonce is None:
                if self.w3.eth.wait_for_transaction_receipt(max_approve.hash)['status'] == 0:
                    log.error(f'admin_max_approve_target transaction {max_approve.hash.hex()} failed')
                    raise SystemExit()

        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return max_approve

    # adminPullAllFunds(erc20s address[])
    def admin_pull_all_funds(self, erc20s, nonce=None, gas=3000000, gas_price=None):
        """this function pulls all funds from the contract
        
        :param erc20s: a list of erc20 addresses
        :type erc20s: list
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce
        :type nonce: int, optional
        :param gas: gas limit of the transaction, defaults to a value of 3000000
        :type gas: int, optional
        :param gas_price: gas price of the transaction, defaults to the gas price of the chain
        :type gas_price: int, optional
        :return: the transaction object of the pull transaction, returns None if the transaction fails
        :rtype: dict, None
        """

        if nonce is None:
            txn_nonce = self.w3.eth.get_transaction_count(self.wallet)
        else:
            txn_nonce = nonce

        if gas_price is None:
            gas_price = self.w3.eth.gas_price

        txn = {'chainId': self.chain, 'gas' : gas, 'gasPrice': gas_price, 'nonce': txn_nonce}

        try:
            pull_all_funds = self.contract.functions.adminPullAllFunds(erc20s).build_transaction(txn)
            pull_all_funds = self.w3.eth.account.sign_transaction(pull_all_funds, self.key)
            self.w3.eth.send_raw_transaction(pull_all_funds.rawTransaction)

            # if a user is not providing a nonce, wait for the transaction to either be confirmed or rejected before continuing
            if nonce is None:
                if self.w3.eth.wait_for_transaction_receipt(pull_all_funds.hash)['status'] == 0:
                    log.error(f'admin_pull_all_funds transaction {pull_all_funds.hash.hex()} failed')
                    raise SystemExit()

        except Exception as e:
            log.error(e, exc_info=True)
            return None

    # adminRebalanceFunds(assetToSell (address), amountToSell (uint256), assetToTarget (address))
    def admin_rebalance_funds(self, asset_to_sell, amount_to_sell, asset_to_target, nonce=None, gas=3000000, gas_price=None):
        """this function rebalances funds from one asset to another on the RubiconMarket

        :param asset_to_sell: the address of the asset to sell
        :type asset_to_sell: str
        :param amount_to_sell: the amount of the asset to sell
        :type amount_to_sell: int
        :param asset_to_target: the address of the asset to target
        :type asset_to_target: str
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce
        :type nonce: int, optional
        :param gas: gas limit of the transaction, defaults to a value of 3000000
        :type gas: int, optional
        :param gas_price: gas price of the transaction, defaults to the gas price of the chain
        :type gas_price: int, optional
        :return: the transaction object of the rebalance transaction, returns None if the transaction fails
        :rtype: dict, None
        """

        if nonce is None:
            txn_nonce = self.w3.eth.get_transaction_count(self.wallet)
        else:
            txn_nonce = nonce

        if gas_price is None:
            gas_price = self.w3.eth.gas_price

        txn = {'chainId': self.chain, 'gas' : gas, 'gasPrice': gas_price, 'nonce': txn_nonce}

        try:
            rebalance = self.contract.functions.adminRebalanceFunds(asset_to_sell, amount_to_sell, asset_to_target).build_transaction(txn)
            rebalance = self.w3.eth.account.sign_transaction(rebalance, self.key)
            self.w3.eth.send_raw_transaction(rebalance.rawTransaction)

            # if a user is not providing a nonce, wait for the transaction to either be confirmed or rejected before continuing
            if nonce is None:
                if self.w3.eth.wait_for_transaction_receipt(rebalance.hash)['status'] == 0:
                    log.error(f'admin_rebalance_funds transaction {rebalance.hash.hex()} failed')
                    raise SystemExit()

        except ValueError:
            log.warning('most likely a checksum error... retrying with checksummed addresses')
            rebalance = self.contract.functions.adminRebalanceFunds(self.w3.to_checksum_address(asset_to_sell), amount_to_sell, self.w3.to_checksum_address(asset_to_target)).build_transaction(txn)
            rebalance = self.w3.eth.account.sign_transaction(rebalance, self.key)
            self.w3.eth.send_raw_transaction(rebalance.rawTransaction)

            # if a user is not providing a nonce, wait for the transaction to either be confirmed or rejected before continuing
            if nonce is None:
                if self.w3.eth.wait_for_transaction_receipt(rebalance.hash)['status'] == 0:
                    log.error(f'admin_rebalance_funds transaction {rebalance.hash.hex()} failed')
                    raise SystemExit()

        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return rebalance

    # approveStrategist(strategist (address))
    def approve_strategist(self, strategist, nonce=None, gas=3000000, gas_price=None):
        """this function approves a strategist to use the aid contract instance

        :param strategist: the address of the strategist to approve
        :type strategist: str
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce
        :type nonce: int, optional
        :param gas: gas limit of the transaction, defaults to a value of 3000000
        :type gas: int, optional
        :param gas_price: gas price of the transaction, defaults to the gas price of the chain
        :type gas_price: int, optional
        :return: the transaction object of the approve transaction, returns None if the transaction fails
        :rtype: dict, None
        """

        if nonce is None:
            txn_nonce = self.w3.eth.get_transaction_count(self.wallet)
        else:
            txn_nonce = nonce

        if gas_price is None:
            gas_price = self.w3.eth.gas_price

        txn = {'chainId': self.chain, 'gas' : gas, 'gasPrice': gas_price, 'nonce': txn_nonce}

        try:
            approve = self.contract.functions.approveStrategist(strategist).build_transaction(txn)
            approve = self.w3.eth.account.sign_transaction(approve, self.key)
            self.w3.eth.send_raw_transaction(approve.rawTransaction)

            # if a user is not providing a nonce, wait for the transaction to either be confirmed or rejected before continuing
            if nonce is None:
                if self.w3.eth.wait_for_transaction_receipt(approve.hash)['status'] == 0:
                    log.error(f'approve_strategist transaction {approve.hash.hex()} failed')
                    raise SystemExit()

        except ValueError:
            log.warning('most likely a checksum error... retrying with checksummed addresses')
            approve = self.contract.functions.approveStrategist(self.w3.to_checksum_address(strategist)).build_transaction(txn)
            approve = self.w3.eth.account.sign_transaction(approve, self.key)
            self.w3.eth.send_raw_transaction(approve.rawTransaction)

            # if a user is not providing a nonce, wait for the transaction to either be confirmed or rejected before continuing
            if nonce is None:
                if self.w3.eth.wait_for_transaction_receipt(approve.hash)['status'] == 0:
                    log.error(f'approve_strategist transaction {approve.hash.hex()} failed')
                    raise SystemExit()

        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return approve

    # batchMarketMakingTrades(tokenPairs (address[2]), askNumerators (uint256[]), askDenominators (uint256[]), bidNumerators (uint256[]), bidDenominators (uint256[]))
    def batch_market_making_trades(self, token_pairs, ask_numerators, ask_denominators, bid_numerators, bid_denominators, nonce=None, gas=3000000, gas_price=None):
        """this function executes a batch of market making trades on the RubiconMarket

        :param token_pairs: the token pairs to trade [token0, token1]
        :type token_pairs: list
        :param ask_numerators: the numerators of the ask prices
        :type ask_numerators: list
        :param ask_denominators: the denominators of the ask prices
        :type ask_denominators: list
        :param bid_numerators: the numerators of the bid prices
        :type bid_numerators: list
        :param bid_denominators: the denominators of the bid prices
        :type bid_denominators: list
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce
        :type nonce: int, optional
        :param gas: gas limit of the transaction, defaults to a value of 3000000
        :type gas: int, optional
        :param gas_price: gas price of the transaction, defaults to the gas price of the chain
        :type gas_price: int, optional
        :return: the transaction object of the batch market making trades transaction, returns None if the transaction fails
        :rtype: dict, None
        """

        if nonce is None:
            txn_nonce = self.w3.eth.get_transaction_count(self.wallet)
        else:
            txn_nonce = nonce

        if gas_price is None:
            gas_price = self.w3.eth.gas_price

        txn = {'chainId': self.chain, 'gas' : gas, 'gasPrice': gas_price, 'nonce': txn_nonce}

        try:
            batch = self.contract.functions.batchMarketMakingTrades(token_pairs, ask_numerators, ask_denominators, bid_numerators, bid_denominators).build_transaction(txn)
            batch = self.w3.eth.account.sign_transaction(batch, self.key)
            self.w3.eth.send_raw_transaction(batch.rawTransaction)

            # if a user is not providing a nonce, wait for the transaction to either be confirmed or rejected before continuing
            if nonce is None:
                if self.w3.eth.wait_for_transaction_receipt(batch.hash)['status'] == 0:
                    log.error(f'batch_market_making_trades transaction {batch.hash.hex()} failed')
                    raise SystemExit()

        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return batch

    # batchRequoteAllOffers(tokenPair (address[2]), askNumerators (uint256[]), askDenominators (uint256[]), bidNumerators (uint256[]), bidDenominators (uint256[]))
    def batch_requote_all_offers(self, token_pair, ask_numerators, ask_denominators, bid_numerators, bid_denominators, nonce=None, gas=3000000, gas_price=None):
        """this function executes a batch requote while clearing all offers the strategist has on the RubiconMarket

        :param token_pair: the token pair to trade [token0, token1]
        :type token_pair: list
        :param ask_numerators: the numerators of the ask prices
        :type ask_numerators: list
        :param ask_denominators: the denominators of the ask prices
        :type ask_denominators: list
        :param bid_numerators: the numerators of the bid prices
        :type bid_numerators: list
        :param bid_denominators: the denominators of the bid prices
        :type bid_denominators: list
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce
        :type nonce: int, optional
        :param gas: gas limit of the transaction, defaults to a value of 3000000
        :type gas: int, optional
        :param gas_price: gas price of the transaction, defaults to the gas price of the chain
        :type gas_price: int, optional
        :return: the transaction object of the batch requote all offers transaction, returns None if the transaction fails
        :rtype: dict, None
        """

        if nonce is None:
            txn_nonce = self.w3.eth.get_transaction_count(self.wallet)
        else:
            txn_nonce = nonce

        if gas_price is None:
            gas_price = self.w3.eth.gas_price

        txn = {'chainId': self.chain, 'gas' : gas, 'gasPrice': gas_price, 'nonce': txn_nonce}

        try:
            batch = self.contract.functions.batchRequoteAllOffers(token_pair, ask_numerators, ask_denominators, bid_numerators, bid_denominators).build_transaction(txn)
            batch = self.w3.eth.account.sign_transaction(batch, self.key)
            self.w3.eth.send_raw_transaction(batch.rawTransaction)

            # if a user is not providing a nonce, wait for the transaction to either be confirmed or rejected before continuing
            if nonce is None:
                if self.w3.eth.wait_for_transaction_receipt(batch.hash)['status'] == 0:
                    log.error(f'batch_requote_all_offers transaction {batch.hash.hex()} failed')
                    raise SystemExit()

        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return batch

    # batchRequoteOffers(ids (uint256[]), tokenPair (address[2]), askNumerators (uint256[]), askDenominators (uint256[]), bidNumerators (uint256[]), bidDenominators (uint256[]))
    def batch_requote_offers(self, ids, token_pair, ask_numerators, ask_denominators, bid_numerators, bid_denominators, nonce=None, gas=3000000, gas_price=None):
        """this function executes a batch requote of all offers that are provided in the ids array
        
        :param ids: the ids of the offers to requote
        :type ids: list
        :param token_pair: the token pair to trade [token0, token1]
        :type token_pair: list
        :param ask_numerators: the numerators of the ask prices
        :type ask_numerators: list
        :param ask_denominators: the denominators of the ask prices
        :type ask_denominators: list
        :param bid_numerators: the numerators of the bid prices
        :type bid_numerators: list
        :param bid_denominators: the denominators of the bid prices
        :type bid_denominators: list
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce
        :type nonce: int, optional
        :param gas: gas limit of the transaction, defaults to a value of 3000000
        :type gas: int, optional
        :param gas_price: gas price of the transaction, defaults to the gas price of the chain
        :type gas_price: int, optional
        :return: the transaction object of the batch requote offers transaction, returns None if the transaction fails
        :rtype: dict, None
        """

        if nonce is None:
            txn_nonce = self.w3.eth.get_transaction_count(self.wallet)
        else:
            txn_nonce = nonce

        if gas_price is None:
            gas_price = self.w3.eth.gas_price

        txn = {'chainId': self.chain, 'gas' : gas, 'gasPrice': gas_price, 'nonce': txn_nonce}

        try:
            batch = self.contract.functions.batchRequoteOffers(ids, token_pair, ask_numerators, ask_denominators, bid_numerators, bid_denominators).build_transaction(txn)
            batch = self.w3.eth.account.sign_transaction(batch, self.key)
            self.w3.eth.send_raw_transaction(batch.rawTransaction)

            # if a user is not providing a nonce, wait for the transaction to either be confirmed or rejected before continuing
            if nonce is None:
                if self.w3.eth.wait_for_transaction_receipt(batch.hash)['status'] == 0:
                    log.error(f'batch_requote_offers transaction {batch.hash.hex()} failed')
                    raise SystemExit()

        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return batch

    # placeMarketMakingTrades(tokenPair (address[2]), askNumerator (uint256), askDenominator (uint256), bidNumerator (uint256), bidDenominator (uint256))
    # aid.batch_market_making_trades([weth.address, usdc.address], [the amount of the asset you will sell], [the amount of the quote you will receive], [the amount of quote you will pay], [the amount of asset you would receive])
    def place_market_making_trades(self, token_pair, ask_numerator, ask_denominator, bid_numerator, bid_denominator, nonce=None, gas=3000000, gas_price=None):
        """this function executes a market making trade on the RubiconMarket
        
        :param token_pair: the token pair to trade [token0, token1]
        :type token_pair: list
        :param ask_numerator: the numerator of the ask price. this is the amount of the asset you will sell
        :type ask_numerator: int
        :param ask_denominator: the denominator of the ask price. this is the amount of the quote you will receive
        :type ask_denominator: int
        :param bid_numerator: the numerator of the bid price. this is the amount of quote you will pay
        :type bid_numerator: int
        :param bid_denominator: the denominator of the bid price. this is the amount of asset you would receive
        :type bid_denominator: int
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce
        :type nonce: int, optional
        :param gas: gas limit of the transaction, defaults to a value of 3000000
        :type gas: int, optional
        :param gas_price: gas price of the transaction, defaults to the gas price of the chain
        :type gas_price: int, optional
        :return: the transaction object of the place market making trades transaction, returns None if the transaction fails
        :rtype: dict, None
        """

        if nonce is None:
            txn_nonce = self.w3.eth.get_transaction_count(self.wallet)
        else:
            txn_nonce = nonce

        if gas_price is None:
            gas_price = self.w3.eth.gas_price

        txn = {'chainId': self.chain, 'gas' : gas, 'gasPrice': gas_price, 'nonce': txn_nonce}

        try:
            trade = self.contract.functions.placeMarketMakingTrades(token_pair, ask_numerator, ask_denominator, bid_numerator, bid_denominator).build_transaction(txn)
            trade = self.w3.eth.account.sign_transaction(trade, self.key)
            self.w3.eth.send_raw_transaction(trade.rawTransaction)

            # if a user is not providing a nonce, wait for the transaction to either be confirmed or rejected before continuing
            if nonce is None:
                if self.w3.eth.wait_for_transaction_receipt(trade.hash)['status'] == 0:
                    log.error(f'place_market_making_trades transaction {trade.hash.hex()} failed')
                    raise SystemExit()

        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return trade

    # removeStrategist(strategist (address))
    def remove_strategist(self, strategist, nonce=None, gas=3000000, gas_price=None):
        """this function removes a strategist from the approved strategist list on the market aid contract
        
        :param strategist: the address of the strategist to remove
        :type strategist: str
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce
        :type nonce: int, optional
        :param gas: gas limit of the transaction, defaults to a value of 3000000
        :type gas: int, optional
        :param gas_price: gas price of the transaction, defaults to the gas price of the chain
        :type gas_price: int, optional
        :return: the transaction object of the remove strategist transaction, returns None if the transaction fails
        :rtype: dict, None
        """

        if nonce is None:
            txn_nonce = self.w3.eth.get_transaction_count(self.wallet)
        else:
            txn_nonce = nonce

        if gas_price is None:
            gas_price = self.w3.eth.gas_price

        txn = {'chainId': self.chain, 'gas' : gas, 'gasPrice': gas_price, 'nonce': txn_nonce}

        try:
            remove = self.contract.functions.removeStrategist(strategist).build_transaction(txn)
            remove = self.w3.eth.account.sign_transaction(remove, self.key)
            self.w3.eth.send_raw_transaction(remove.rawTransaction)

            # if a user is not providing a nonce, wait for the transaction to either be confirmed or rejected before continuing
            if nonce is None:
                if self.w3.eth.wait_for_transaction_receipt(remove.hash)['status'] == 0:
                    log.error(f'remove_strategist transaction {remove.hash.hex()} failed')
                    raise SystemExit()

        except ValueError:
            log.warning('most likely a checksum error... retrying with checksummed addresses')
            remove = self.contract.functions.removeStrategist(self.w3.to_checksum_address(strategist)).build_transaction(txn)
            remove = self.w3.eth.account.sign_transaction(remove, self.key)
            self.w3.eth.send_raw_transaction(remove.rawTransaction)

            # if a user is not providing a nonce, wait for the transaction to either be confirmed or rejected before continuing
            if nonce is None:
                if self.w3.eth.wait_for_transaction_receipt(remove.hash)['status'] == 0:
                    log.error(f'remove_strategist transaction {remove.hash.hex()} failed')
                    raise SystemExit()

        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return remove

    # requote(id (uint256), tokenPair (address[2]), askNumerator (uint256), askDenominator (uint256), bidNumerator (uint256), bidDenominator (uint256))
    def requote(self, id, token_pair, ask_numerator, ask_denominator, bid_numerator, bid_denominator, nonce=None, gas=3000000, gas_price=None):
        """this function requotes an offer on the RubiconMarket
        
        :param id: the id of the offer to requote
        :type id: int
        :param token_pair: the token pair to trade [token0, token1]
        :type token_pair: list
        :param ask_numerator: the numerator of the ask price
        :type ask_numerator: int
        :param ask_denominator: the denominator of the ask price
        :type ask_denominator: int
        :param bid_numerator: the numerator of the bid price
        :type bid_numerator: int
        :param bid_denominator: the denominator of the bid price
        :type bid_denominator: int
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce
        :type nonce: int, optional
        :param gas: gas limit of the transaction, defaults to a value of 3000000
        :type gas: int, optional
        :param gas_price: gas price of the transaction, defaults to the gas price of the chain
        :type gas_price: int, optional
        :return: the transaction object of the requote transaction, returns None if the transaction fails
        :rtype: dict, None
        """

        if nonce is None:
            txn_nonce = self.w3.eth.get_transaction_count(self.wallet)
        else:
            txn_nonce = nonce

        if gas_price is None:
            gas_price = self.w3.eth.gas_price

        txn = {'chainId': self.chain, 'gas' : gas, 'gasPrice': gas_price, 'nonce': txn_nonce}

        try:
            requote = self.contract.functions.requote(id, token_pair, ask_numerator, ask_denominator, bid_numerator, bid_denominator).build_transaction(txn)
            requote = self.w3.eth.account.sign_transaction(requote, self.key)
            self.w3.eth.send_raw_transaction(requote.rawTransaction)

            # if a user is not providing a nonce, wait for the transaction to either be confirmed or rejected before continuing
            if nonce is None:
                if self.w3.eth.wait_for_transaction_receipt(requote.hash)['status'] == 0:
                    log.error(f'requote transaction {requote.hash.hex()} failed')
                    raise SystemExit()

        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return requote

    # scrubStrategistTrade(id (uint256))
    def scrub_strategist_trade(self, id, nonce=None, gas=3000000, gas_price=None):
        """this function scrubs a strategist trade from the RubiconMarket contract

        :param id: the id of the trade to scrub
        :type id: int
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce
        :type nonce: int, optional
        :param gas: gas limit of the transaction, defaults to a value of 3000000
        :type gas: int, optional
        :param gas_price: gas price of the transaction, defaults to the gas price of the chain
        :type gas_price: int, optional
        """

        if nonce is None:
            txn_nonce = self.w3.eth.get_transaction_count(self.wallet)
        else:
            txn_nonce = nonce

        if gas_price is None:
            gas_price = self.w3.eth.gas_price

        txn = {'chainId': self.chain, 'gas' : gas, 'gasPrice': gas_price, 'nonce': txn_nonce}

        try:
            scrub = self.contract.functions.scrubStrategistTrade(id).build_transaction(txn)
            scrub = self.w3.eth.account.sign_transaction(scrub, self.key)
            self.w3.eth.send_raw_transaction(scrub.rawTransaction)

            # if a user is not providing a nonce, wait for the transaction to either be confirmed or rejected before continuing
            if nonce is None:
                if self.w3.eth.wait_for_transaction_receipt(scrub.hash)['status'] == 0:
                    log.error(f'scrub_strategist_trade transaction {scrub.hash.hex()} failed')
                    raise SystemExit()

        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return scrub

    # scrubStrategistTrades(ids (uint256[]))
    def scrub_strategist_trades(self, ids, nonce=None, gas=3000000, gas_price=None):
        """this function scrubs a list of strategist trades from the RubiconMarket contract

        :param ids: the ids of the trades to scrub
        :type ids: list
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce
        :type nonce: int, optional
        :param gas: gas limit of the transaction, defaults to a value of 3000000
        :type gas: int, optional
        :param gas_price: gas price of the transaction, defaults to the gas price of the chain
        :type gas_price: int, optional
        :return: the transaction object of the scrub transaction, returns None if the transaction fails
        :rtype: dict, None
        """

        if nonce is None:
            txn_nonce = self.w3.eth.get_transaction_count(self.wallet)
        else:
            txn_nonce = nonce

        if gas_price is None:
            gas_price = self.w3.eth.gas_price

        txn = {'chainId': self.chain, 'gas' : gas, 'gasPrice': gas_price, 'nonce': txn_nonce}

        try:
            scrub = self.contract.functions.scrubStrategistTrades(ids).build_transaction(txn)
            scrub = self.w3.eth.account.sign_transaction(scrub, self.key)
            self.w3.eth.send_raw_transaction(scrub.rawTransaction)

            # if a user is not providing a nonce, wait for the transaction to either be confirmed or rejected before continuing
            if nonce is None:
                if self.w3.eth.wait_for_transaction_receipt(scrub.hash)['status'] == 0:
                    log.error(f'scrub_strategist_trades transaction {scrub.hash.hex()} failed')
                    raise SystemExit()

        except Exception as e:
            log.error(e, exc_info=True)
            return None

        return scrub


from _decimal import Decimal

from rubi import ERC20


class Pair:
    """Class representing a trading asset pair, e.g. WETH/USDC would have WETH as the base asset and USDC as the quote
    asset.

    :param name: The name of the pair, e.g. WETH/USDC.
    :type name: str
    :param base_asset: Base asset of the pair.
    :type base_asset: ERC20
    :param quote_asset: Quote asset of the pair.
    :type quote_asset: ERC20
    :param current_base_asset_allowance: The base asset spending allowance of the RubiconMarket contract.
    :type current_base_asset_allowance: Decimal
    :param current_quote_asset_allowance: The quote asset spending allowance of the RubiconMarket contract.
    :type current_quote_asset_allowance: Decimal
    """

    def __init__(
        self,
        name: str,
        base_asset: ERC20,
        quote_asset: ERC20,
        current_base_asset_allowance: Decimal,
        current_quote_asset_allowance: Decimal
    ):
        self.name = name

        self.base_asset: ERC20 = base_asset
        self.quote_asset: ERC20 = quote_asset

        self.bid_identifier: str = base_asset.w3.solidity_keccak(
            abi_types=["address", "address"],
            values=[self.quote_asset.address, self.base_asset.address]
        ).hex()
        self.ask_identifier: str = base_asset.w3.solidity_keccak(
            abi_types=["address", "address"],
            values=[self.base_asset.address, self.quote_asset.address]
        ).hex()

        # TODO: think about structure of allowances on this class. Currently this only caters for the RubiconMarket
        #  contract.
        self.current_base_asset_allowance = current_base_asset_allowance
        self.current_quote_asset_allowance = current_quote_asset_allowance

    def update_base_asset_allowance(self, new_base_asset_allowance: Decimal) -> None:
        """Update the current base asset allowance.

        :param new_base_asset_allowance: New base asset allowance.
        :type new_base_asset_allowance: Decimal
        :return: None
        """
        self.current_base_asset_allowance = new_base_asset_allowance

    def update_quote_asset_allowance(self, new_quote_asset_allowance: Decimal) -> None:
        """Update the current quote asset allowance.

        :param new_quote_asset_allowance: New quote asset allowance.
        :type new_quote_asset_allowance: Decimal
        :return: None
        """
        self.current_quote_asset_allowance = new_quote_asset_allowance


class PairDoesNotExistException(Exception):
    """Exception raised when an asset pair does not exist.

    :param msg: Error message.
    :type msg: str
    """

    def __init__(self, msg: str):
        super().__init__(msg)



from subgrounds import Subgrounds
from subgrounds.pagination import ShallowStrategy

from rubi.data.sources.helper import networks

class MarketData: 
    """this class acts as an access point to a variety of data from the RubiconMarket.sol contract. it acts as a data processing layer built using the subgrounds library and the subgraphs maintained at the follwing repo: https://github.com/RubiconDeFi/rubi-subgraphs
    """

    def __init__(self, subgrounds, chain_id):
        """constructor method
        
        :param subgrounds: the subgrounds object
        :type subgrounds: Subgrounds
        :param chain_id: the chain id of the network that is of interest
        :type chain_id: int
        """
        self.network = networks[chain_id]()
        self.subgrounds = subgrounds
        self.rubicon_market_light = self.subgrounds.load_subgraph(self.network.rubicon_market_light)
        self.boiler_plate = self.subgrounds.load_subgraph(self.network.boiler_plate)

    ######################################################################
    # data collection 
    ######################################################################

    def get_offers(self, maker = None, pair = None, filled = None, cancelled = None, live = None, pay_gem = None, buy_gem = None, start_time = None, end_time = None, first = 1000000000): 
        """returns a dataframe of all offers placed on the market contract. allows for the following filters: 

        :param maker: the address of the maker of the offer, defaults to None
        :type maker: str, optional
        :param pair: the address of the pair of the offer, defaults to None. direction of the pair matters and corresponds to the array passed in the following order: [pay_gem, buy_gem]. defaults to None
        :type pair: array, optional
        :param filled: whether the offer has been filled, defaults to None
        :type filled: bool, optional
        :param cancelled: whether the offer has been cancelled, defaults to None
        :type cancelled: bool, optional
        :param live: whether the offer is live, defaults to None
        :type live: bool, optional
        :param pay_gem: the address of the pay_gem of the offer, defaults to None
        :type pay_gem: str, optional
        :param buy_gem: the address of the buy_gem of the offer, defaults to None
        :type buy_gem: str, optional
        :param start_time: the start time of the offer, defaults to None. in unix time
        :type start_time: int, optional
        :param end_time: the end time of the offer, defaults to None. in unix time
        :type end_time: int, optional
        """

        # set the offer entity 
        Offer = self.rubicon_market_light.Offer

        # create the synthetic fields for the offer entity
        Offer.pay_amt_formatted = Offer.pay_amt / 10 ** Offer.pay_gem.decimals
        Offer.buy_amt_formatted = Offer.buy_amt / 10 ** Offer.buy_gem.decimals
        Offer.paid_amt_formatted = Offer.paid_amt / 10 ** Offer.pay_gem.decimals
        Offer.bought_amt_formatted = Offer.bought_amt / 10 ** Offer.buy_gem.decimals

        # for each of the filters, add the filter to the where clause
        # TODO: there is most likely a more elegant way to do the lower case conversion of the variables, probably using a decorator?
        where = []
        if maker:
            maker = maker.lower()
            where.append(Offer.maker == maker)
        # TODO: this should be modified to return all offers for a given pair, not just the given direction
        if pair:
            pair[0] = pair[0].lower()
            pair[1] = pair[1].lower()
            where.append(Offer.pay_gem == pair[0])
            where.append(Offer.buy_gem == pair[1])
        if filled:
            where.append(Offer.filled == filled)
        if cancelled:
            where.append(Offer.cancelled == cancelled)
        if live:
            where.append(Offer.live == live)
        if pay_gem:
            pay_gem = pay_gem.lower()
            where.append(Offer.pay_gem == pay_gem)
        if buy_gem:
            buy_gem = buy_gem.lower()
            where.append(Offer.buy_gem == buy_gem)
        if start_time:
            where.append(Offer.timestamp >= start_time)
        if end_time:
            where.append(Offer.timestamp <= end_time)

        # TODO: this is a current limit of the subgrounds library, it does not support querying all as a funtion option, so a large number must be provided
        # TODO: help a frend, this would be a good pr to add to the subgrounds library :)
        if where == []:
            offers = self.rubicon_market_light.Query.offers(first = first)
        else:
            offers = self.rubicon_market_light.Query.offers(first = first, where = where)

        field_paths = [
            offers.id,
            offers.pay_gem.symbol,
            offers.pay_amt_formatted,
            offers.paid_amt_formatted,
            offers.buy_gem.symbol,
            offers.buy_amt_formatted,
            offers.bought_amt_formatted,
            offers.maker.id,
            offers.transaction.id,
            offers.transaction.timestamp,
            offers.pay_gem.id,
            offers.buy_gem.id,
            offers.pay_amt,
            offers.buy_amt,
            offers.paid_amt, 
            offers.bought_amt
        ]  

        df = self.subgrounds.query_df(field_paths, pagination_strategy=ShallowStrategy) 

        return df

    def get_filled_offers(self, maker = None, taker = None, pair = None, filled = None, cancelled = None, live = None, pay_gem = None, buy_gem = None, start_time = None, end_time = None, first = 1000000000):
        """returns a dataframe of offers that have received fill with each row corresponding to the take event on the offer. allows for the following filters:

        :param maker: the address of the maker of the offer, defaults to None
        :type maker: str, optional
        :param taker: the address of the taker of the offer, defaults to None
        :type taker: str, optional
        :param pair: the address of the pair of the offer, defaults to None. direction of the pair matters and corresponds to the array passed in the following order: [pay_gem, buy_gem]. defaults to None
        :type pair: array, optional
        :param filled: whether the offer has been filled, defaults to None
        :type filled: bool, optional
        :param cancelled: whether the offer has been cancelled, defaults to None
        :type cancelled: bool, optional
        :param live: whether the offer is live, defaults to None
        :type live: bool, optional
        :param pay_gem: the address of the pay_gem of the offer, defaults to None
        :type pay_gem: str, optional
        :param buy_gem: the address of the buy_gem of the offer, defaults to None
        :type buy_gem: str, optional
        :param start_time: the start time of the offer, defaults to None. in unix time
        :type start_time: int, optional
        :param end_time: the end time of the offer, defaults to None. in unix time
        :type end_time: int, optional
        """

        # set the offer entity 
        Offer = self.rubicon_market_light.Offer
        Take = self.rubicon_market_light.Take

        # create the synthetic fields for the offer entity
        Offer.pay_amt_formatted = Offer.pay_amt / 10 ** Offer.pay_gem.decimals
        Offer.buy_amt_formatted = Offer.buy_amt / 10 ** Offer.buy_gem.decimals
        Offer.paid_amt_formatted = Offer.paid_amt / 10 ** Offer.pay_gem.decimals
        Offer.bought_amt_formatted = Offer.bought_amt / 10 ** Offer.buy_gem.decimals

        Take.pay_amt_formatted = Take.pay_amt / 10 ** Take.pay_gem.decimals
        Take.buy_amt_formatted = Take.buy_amt / 10 ** Take.buy_gem.decimals

        # for each of the filters, add the filter to the where clause
        # TODO: there is most likely a more elegant way to do the lower case conversion of the variables, probably using a decorator?
        where = []
        if maker:
            maker = maker.lower()
            where.append(Offer.maker == maker)
        # TODO: this should be modified to return all offers for a given pair, not just the given direction
        if taker: 
            taker = taker.lower()
            where.append(Offer.takes.taker.id == taker)
        if pair:
            pair[0] = pair[0].lower()
            pair[1] = pair[1].lower()
            where.append(Offer.pay_gem == pair[0])
            where.append(Offer.buy_gem == pair[1])
        if filled:
            where.append(Offer.filled == filled)
        if cancelled:
            where.append(Offer.cancelled == cancelled)
        if live:
            where.append(Offer.live == live)
        if pay_gem:
            pay_gem = pay_gem.lower()
            where.append(Offer.pay_gem == pay_gem)
        if buy_gem:
            buy_gem = buy_gem.lower()
            where.append(Offer.buy_gem == buy_gem)
        if start_time:
            where.append(Offer.timestamp >= start_time)
        if end_time:
            where.append(Offer.timestamp <= end_time)

        if where == []:
            offers = self.rubicon_market_light.Query.offers(first = first)
        else:
            offers = self.rubicon_market_light.Query.offers(first = first, where = where)

        field_paths = [
            offers.id,
            offers.pay_gem.symbol,
            offers.pay_amt_formatted,
            offers.paid_amt_formatted,
            offers.buy_gem.symbol,
            offers.buy_amt_formatted,
            offers.bought_amt_formatted,
            offers.maker.id,
            offers.transaction.id,
            offers.transaction.timestamp,
            offers.pay_gem.id,
            offers.buy_gem.id,
            offers.pay_amt,
            offers.buy_amt,
            offers.paid_amt, 
            offers.bought_amt,

            offers.takes.pay_gem.symbol,
            offers.takes.pay_amt_formatted,
            offers.takes.buy_gem.symbol,
            offers.takes.buy_amt_formatted,
            offers.takes.taker.id,
            offers.takes.transaction.id,
            offers.takes.timestamp, 
            offers.takes.pay_gem.id,
            offers.takes.buy_gem.id,
            offers.takes.pay_amt,
            offers.takes.buy_amt
        ]

        df = self.subgrounds.query_df(field_paths, pagination_strategy=ShallowStrategy) 

        return df


    def get_trades(self, taker = None, maker = None, pair = None, pay_gem = None, buy_gem = None, start_time = None, end_time = None, first = 1000000000): 
        """returns a dataframe of all trades on the market contract. allows for the following filters: 

        :param taker: the address of the taker of the trade, defaults to None
        :type taker: str, optional
        :param maker: the address of the maker of the offer, defaults to None
        :type maker: str, optional
        :param pair: the address of the pair of the offer, defaults to None. direction of the pair matters and corresponds to the array passed in the following order: [pay_gem, buy_gem]. defaults to None
        :type pair: array, optional
        :param filled: whether the offer has been filled, defaults to None
        :type filled: bool, optional
        :param cancelled: whether the offer has been cancelled, defaults to None
        :type cancelled: bool, optional
        :param live: whether the offer is live, defaults to None
        :type live: bool, optional
        :param pay_gem: the address of the pay_gem of the offer, defaults to None
        :type pay_gem: str, optional
        :param buy_gem: the address of the buy_gem of the offer, defaults to None
        :type buy_gem: str, optional
        :param start_time: the start time of the offer, defaults to None. in unix time
        :type start_time: int, optional
        :param end_time: the end time of the offer, defaults to None. in unix time
        :type end_time: int, optional
        """

        # set the trade entity 
        Take = self.rubicon_market_light.Take

        # create the synthetic fields for the trade entity
        Take.pay_amt_formatted = Take.pay_amt / 10 ** Take.pay_gem.decimals
        Take.buy_amt_formatted = Take.buy_amt / 10 ** Take.buy_gem.decimals

        # for each of the filters, add the filter to the where clause
        where = []
        if taker:
            taker = taker.lower()
            where.append(Take.taker == taker)
        if maker:
            maker = maker.lower()
            where.append(Take.maker == maker)
        # TODO: this should be modified to return all offers for a given pair, not just the given direction
        if pair:
            pair[0] = pair[0].lower()
            pair[1] = pair[1].lower()
            where.append(Take.pay_gem == pair[0])
            where.append(Take.buy_gem == pair[1])
        if pay_gem:
            pay_gem = pay_gem.lower()
            where.append(Take.pay_gem == pay_gem)
        if buy_gem:
            buy_gem = buy_gem.lower()
            where.append(Take.buy_gem == buy_gem)
        if start_time:
            where.append(Take.timestamp >= start_time)
        if end_time:
            where.append(Take.timestamp <= end_time)

        # TODO: this is a current limit of the subgrounds library, it does not support querying all as a funtion option, so a large number must be provided
        # TODO: help a frend, this would be a good pr to add to the subgrounds library :)
        if where == []:
            takes = self.rubicon_market_light.Query.takes(first = first)
        else:
            takes = self.rubicon_market_light.Query.takes(first = first, where = where)

        field_paths = [
            takes.id,
            takes.pay_gem.symbol,
            takes.pay_amt_formatted,
            takes.buy_gem.symbol,
            takes.buy_amt_formatted,
            takes.taker.id,
            takes.transaction.id,
            takes.transaction.timestamp,
            takes.pay_gem.id,
            takes.buy_gem.id,
            takes.pay_amt,
            takes.buy_amt,
            takes.offer.id, 
            takes.offer.maker.id
        ]  

        df = self.subgrounds.query_df(field_paths, pagination_strategy=ShallowStrategy) 

        return df

    def get_detailed_offers(self, maker = None, pair = None, filled = None, cancelled = None, live = None, pay_gem = None, buy_gem = None, start_time = None, end_time = None, first = 1000000000): 
            """returns a dataframe of all offers placed on the market contract with price data. allows for the following filters:
            
            :param maker: the address of the maker of the offer, defaults to None
            :type maker: str, optional
            :param pair: the address of the pair of the offer, defaults to None. direction of the pair matters and corresponds to the array passed in the following order: [pay_gem, buy_gem]. defaults to None
            :type pair: array, optional
            :param filled: whether the offer has been filled, defaults to None
            :type filled: bool, optional
            :param cancelled: whether the offer has been cancelled, defaults to None
            :type cancelled: bool, optional
            :param live: whether the offer is live, defaults to None
            :type live: bool, optional
            :param pay_gem: the address of the pay_gem of the offer, defaults to None
            :type pay_gem: str, optional
            :param buy_gem: the address of the buy_gem of the offer, defaults to None
            :type buy_gem: str, optional
            :param start_time: the start time of the offer, defaults to None. in unix time
            :type start_time: int, optional
            :param end_time: the end time of the offer, defaults to None. in unix time
            :type end_time: int, optional
            """

            # set the offer entity 
            Offer = self.boiler_plate.Offer

            # for each of the filters, add the filter to the where clause
            # TODO: there is most likely a more elegant way to do the lower case conversion of the variables, probably using a decorator?
            where = []
            if maker:
                maker = maker.lower()
                where.append(Offer.maker == maker)
            # TODO: this should be modified to return all offers for a given pair, not just the given direction
            if pair:
                pair[0] = pair[0].lower()
                pair[1] = pair[1].lower()
                where.append(Offer.pay_gem == pair[0])
                where.append(Offer.buy_gem == pair[1])
            if filled:
                where.append(Offer.filled == filled)
            if cancelled:
                where.append(Offer.cancelled == cancelled)
            if live:
                where.append(Offer.live == live)
            if pay_gem:
                pay_gem = pay_gem.lower()
                where.append(Offer.pay_gem == pay_gem)
            if buy_gem:
                buy_gem = buy_gem.lower()
                where.append(Offer.buy_gem == buy_gem)
            if start_time:
                where.append(Offer.timestamp >= start_time)
            if end_time:
                where.append(Offer.timestamp <= end_time)

            if where == []:
                offers = self.boiler_plate.Query.offers(first = first)
            else:
                offers = self.boiler_plate.Query.offers(first = first, where = where)

            field_paths = [
                offers.id,
                offers.pay_gem.id,
                offers.pay_amt_formatted,
                offers.pay_amt_usd,
                offers.buy_gem.id,
                offers.buy_amt_formatted,
                offers.buy_amt_usd,
                offers.bought_amt_formatted,
                offers.maker.id,
                offers.transaction.id,
                offers.transaction.timestamp,
                offers.pay_amt,
                offers.buy_amt,
                offers.paid_amt, 
                offers.paid_amt_formatted,
                offers.paid_amt_usd,
                offers.bought_amt,
                offers.bought_amt_formatted,
                offers.bought_amt_usd,
                offers.filled,
                offers.cancelled,
                offers.live,
                offers.removed_timestamp
            ]  

            df = self.subgrounds.query_df(field_paths, pagination_strategy=ShallowStrategy) 

            return df

    def get_detailed_trades(self, taker = None, maker = None, pair = None, pay_gem = None, buy_gem = None, start_time = None, end_time = None, first = 1000000000): 
            """returns a dataframe of all trades on the market contract with price data. allows for the following filters: 

            :param taker: the address of the taker of the trade, defaults to None
            :type taker: str, optional
            :param maker: the address of the maker of the offer, defaults to None
            :type maker: str, optional
            :param pair: the address of the pair of the offer, defaults to None. direction of the pair matters and corresponds to the array passed in the following order: [pay_gem, buy_gem]. defaults to None
            :type pair: array, optional
            :param filled: whether the offer has been filled, defaults to None
            :type filled: bool, optional
            :param cancelled: whether the offer has been cancelled, defaults to None
            :type cancelled: bool, optional
            :param live: whether the offer is live, defaults to None
            :type live: bool, optional
            :param pay_gem: the address of the pay_gem of the offer, defaults to None
            :type pay_gem: str, optional
            :param buy_gem: the address of the buy_gem of the offer, defaults to None
            :type buy_gem: str, optional
            :param start_time: the start time of the offer, defaults to None. in unix time
            :type start_time: int, optional
            :param end_time: the end time of the offer, defaults to None. in unix time
            :type end_time: int, optional
            """

            # set the trade entity 
            Take = self.boiler_plate.Take

            # for each of the filters, add the filter to the where clause
            where = []
            if taker:
                taker = taker.lower()
                where.append(Take.taker == taker)
            if maker:
                maker = maker.lower()
                where.append(Take.maker == maker)
            # TODO: this should be modified to return all offers for a given pair, not just the given direction
            if pair:
                pair[0] = pair[0].lower()
                pair[1] = pair[1].lower()
                where.append(Take.pay_gem == pair[0])
                where.append(Take.buy_gem == pair[1])
            if pay_gem:
                pay_gem = pay_gem.lower()
                where.append(Take.pay_gem == pay_gem)
            if buy_gem:
                buy_gem = buy_gem.lower()
                where.append(Take.buy_gem == buy_gem)
            if start_time:
                where.append(Take.timestamp >= start_time)
            if end_time:
                where.append(Take.timestamp <= end_time)

            # TODO: this is a current limit of the subgrounds library, it does not support querying all as a funtion option, so a large number must be provided
            # TODO: help a frend, this would be a good pr to add to the subgrounds library :)
            if where == []:
                takes = self.boiler_plate.Query.takes(first = first)
            else:
                takes = self.boiler_plate.Query.takes(first = first, where = where)

            field_paths = [
                takes.id,
                takes.pay_gem.id,
                takes.pay_amt_formatted,
                takes.pay_amt_usd,
                takes.pay_gem_price,
                takes.buy_gem.id,
                takes.buy_amt_formatted,
                takes.buy_amt_usd,
                takes.buy_gem_price,
                takes.taker.id,
                takes.transaction.id,
                takes.transaction.timestamp,
                takes.pay_amt,
                takes.buy_amt,
                takes.offer.id, 
                takes.offer.maker.id
            ]  

            df = self.subgrounds.query_df(field_paths, pagination_strategy=ShallowStrategy) 

            return df


import logging as log
from _decimal import Decimal
from multiprocessing import Queue
from threading import Thread
from time import sleep
from typing import Union, List, Optional, Dict, Type, Any, Callable

from eth_typing import ChecksumAddress
from web3.types import EventData

from rubi.contracts import (
    RubiconMarket,
    RubiconRouter,
    ERC20,
)
from rubi.network import (
    Network,
)
from rubi.types import (
    OrderSide,
    NewMarketOrder,
    NewLimitOrder,
    Pair,
    OrderBook,
    PairDoesNotExistException,
    BaseEvent,
    OrderEvent,
    Transaction,
    BaseNewOrder,
    NewCancelOrder,
    UpdateLimitOrder
)


class Client:
    """This class is a client for Rubicon. It aims to provide a simple and understandable interface when interacting
    with the Rubicon protocol. If not instantiated with a wallet and key then all the methods that require signing
    will throw an error.

    :param network: A Network instance
    :type network: Network
    :param message_queue: Optional message queue for processing events (optional, default is None).
    :type message_queue: Optional[Queue]
    :param wallet: Wallet address (optional, default is None).
    :type wallet: Optional[ChecksumAddress]
    :param key: Key for the wallet (optional, default is None).
    :type key: Optional[str]
    """

    def __init__(
        self,
        network: Network,
        message_queue: Optional[Queue] = None,
        wallet: Optional[Union[ChecksumAddress, str]] = None,
        key: Optional[str] = None,
    ):
        """constructor method."""
        self.network = network

        self.wallet = self.network.w3.to_checksum_address(wallet) if wallet else wallet  # type: ChecksumAddress |  None
        self.key = key  # type: str |  None

        self.market = RubiconMarket.from_network(network=self.network, wallet=self.wallet, key=self.key)
        self.router = RubiconRouter.from_network(network=self.network, wallet=self.wallet, key=self.key)

        self._pairs: Dict[str, Pair] = {}
        self._pair_orderbooks: Dict[str, OrderBook] = {}

        self.message_queue = message_queue  # type: Queue | None

    @classmethod
    def from_http_node_url(
        cls,
        http_node_url: str,
        custom_token_addresses_file: Optional[str] = None,
        message_queue: Optional[Queue] = None,
        wallet: Optional[Union[ChecksumAddress, str]] = None,
        key: Optional[str] = None,
    ):
        """Initialize a Client using a http_node_url.

        :param http_node_url: URL of the HTTP node.
        :type http_node_url: str
        :param custom_token_addresses_file: The name of a yaml file (relative to the current working directory) with
            custom token addresses. Overwrites the token config found in network_config/{chain}/network.yaml.
            (optional, default is None).
        :type custom_token_addresses_file: Optional[str]
        :param message_queue: Optional message queue for processing events (optional, default is None).
        :type message_queue: Optional[Queue]
        :param wallet: Wallet address (optional, default is None).
        :type wallet: Optional[Union[ChecksumAddress, str]]
        :param key: Key for the wallet (optional, default is None).
        :type key: str
        """
        network = Network.from_config(
            http_node_url=http_node_url,
            custom_token_addresses_file=custom_token_addresses_file
        )

        return cls(
            network=network,
            message_queue=message_queue,
            wallet=wallet,
            key=key
        )

    ######################################################################
    # pair methods
    ######################################################################

    def add_pair(
        self, pair_name: str,
        base_asset_allowance: Optional[Decimal] = None,
        quote_asset_allowance: Optional[Decimal] = None
    ) -> None:
        """Add a Pair to the Client. This method creates a Pair instance and adds it to the Client's internal
        _pairs dictionary. Additionally, this method updates the spender allowance of the Rubicon Market for both
        base asset and the quote asset.

        :param pair_name: Name of the Pair in the format "<base_asset>/<quote_asset>".
        :type pair_name: str
        :param base_asset_allowance: Allowance for the base asset (optional, default is None).
        :type base_asset_allowance: Optional[Decimal]
        :param quote_asset_allowance: Allowance for the quote asset (optional, default is None).
        :type quote_asset_allowance: Optional[Decimal]
        """

        base, quote = pair_name.split("/")

        base_asset = ERC20.from_network(name=base, network=self.network, wallet=self.wallet, key=self.key)
        quote_asset = ERC20.from_network(name=quote, network=self.network, wallet=self.wallet, key=self.key)

        current_base_asset_allowance = base_asset.to_decimal(
            number=base_asset.allowance(owner=self.wallet, spender=self.market.address)
        )
        current_quote_asset_allowance = quote_asset.to_decimal(
            number=quote_asset.allowance(owner=self.wallet, spender=self.market.address)
        )

        if current_base_asset_allowance == Decimal("0") or current_quote_asset_allowance == Decimal("0"):
            log.warning("allowance for base or quote asset is zero. this may cause issues when placing orders")

        self._pairs[f"{base}/{quote}"] = Pair(
            name=pair_name,
            base_asset=base_asset,
            quote_asset=quote_asset,
            current_base_asset_allowance=current_base_asset_allowance,
            current_quote_asset_allowance=current_quote_asset_allowance
        )

        # only edit allowance if client has signing rights
        if self.wallet is not None and self.key is not None:
            self.update_pair_allowance(
                pair_name=pair_name,
                new_base_asset_allowance=base_asset_allowance,
                new_quote_asset_allowance=quote_asset_allowance
            )

    def get_pairs_list(self) -> List[str]:
        """Get a list of all pair names in the clients internal _pairs dictionary.

        :return: List of pair names.
        :rtype: List[str]
        """
        return list(self._pairs.keys())

    def update_pair_allowance(
        self,
        pair_name: str,
        new_base_asset_allowance: Optional[Decimal] = None,
        new_quote_asset_allowance: Optional[Decimal] = None
    ) -> None:
        """Update the allowance for the base and quote assets of a pair if the current allowance is different from the
        new allowance. This method also updates the Pair data structure so that the allowance can be read without having
        to do a call to the chain.

        :param pair_name: Name of the pair.
        :type pair_name: str
        :param new_base_asset_allowance: New allowance for the base asset. (optional, default is None).
        :type new_base_asset_allowance: Optional[Decimal]
        :param new_quote_asset_allowance: New allowance for the quote asset. (optional, default is None).
        :type new_quote_asset_allowance: Optional[Decimal]
        :raises PairDoesNotExistException: If the pair does not exist in the clients internal _pairs dict.
        """
        pair = self.get_pair(pair_name=pair_name)

        if new_base_asset_allowance and pair.current_base_asset_allowance != new_base_asset_allowance:
            self._update_asset_allowance(
                asset=pair.base_asset,
                spender=self.market.address,
                new_allowance=new_base_asset_allowance
            )
            pair.update_base_asset_allowance(new_base_asset_allowance=new_base_asset_allowance)

        if new_quote_asset_allowance and pair.current_base_asset_allowance != new_quote_asset_allowance:
            self._update_asset_allowance(
                asset=pair.quote_asset,
                spender=self.market.address,
                new_allowance=new_quote_asset_allowance
            )
            pair.update_quote_asset_allowance(new_quote_asset_allowance=new_quote_asset_allowance)

    def get_pair(self, pair_name: str) -> Pair:
        """Retrieves the Pair object associated with the specified pair name. If the pair does not exist
        in the client, it raises a PairDoesNotExistException.

        :param pair_name: Name of the pair.
        :type pair_name: str
        :return: The Pair object.
        :rtype: Pair
        :raises PairDoesNotExistException: If the pair does not exist in the client.
        """
        pair = self._pairs.get(pair_name)

        if pair is None:
            raise PairDoesNotExistException(
                "add pair to the client using the add_pair method before placing orders for the pair"
            )

        return pair

    def remove_pair(self, pair_name: str) -> None:
        """Removes a pair from the client. It updates the pair's asset allowances
        to zero and deletes the pair and its corresponding order book from the client.

        :param pair_name: Name of the pair to remove.
        :type pair_name: str
        :raises PairDoesNotExistException: If the pair does not exist in the client before removal.
        """
        self.update_pair_allowance(
            pair_name=pair_name,
            new_base_asset_allowance=Decimal("0"),
            new_quote_asset_allowance=Decimal("0")
        )

        del self._pairs[pair_name]
        if self._pair_orderbooks.get(pair_name):
            del self._pair_orderbooks[pair_name]

    ######################################################################
    # orderbook methods
    ######################################################################

    def get_orderbook(self, pair_name: str) -> OrderBook:
        """Retrieve the order book for a specific pair from the Rubicon Router.

        :param pair_name: Name of the pair to retrieve the order book for.
        :type pair_name: str
        :return: The order book for the specified pair.
        :rtype: OrderBook
        :raises PairDoesNotExistException: If the pair does not exist in the client.
        """
        pair = self.get_pair(pair_name=pair_name)

        rubicon_offer_book = self.router.get_book_from_pair(
            asset=pair.base_asset.address,
            quote=pair.quote_asset.address
        )

        return OrderBook.from_rubicon_offer_book(
            offer_book=rubicon_offer_book,
            base_asset=pair.base_asset,
            quote_asset=pair.quote_asset
        )

    def start_orderbook_poller(self, pair_name: str, poll_time: int = 2) -> None:
        """Starts a background thread that continuously polls the order book for the specified pair
        at a specified polling interval. The retrieved order book is added to the message queue of the client.
        The poller will run until the pair is removed from the client.

        :param pair_name: Name of the pair to start the order book poller for.
        :type pair_name: str
        :param poll_time: Polling interval in seconds, defaults to 2 seconds.
        :type poll_time: int, optional
        :raises Exception: If the message queue is not configured.
        :raises PairDoesNotExistException: If the pair does not exist in the client.
        """

        if self.message_queue is None:
            raise Exception("orderbook poller is configured to place messages on the message queue. message queue"
                            "cannot be none")

        # Check that pair is defined before starting
        pair = self.get_pair(pair_name=pair_name)

        thread = Thread(
            target=self._start_orderbook_poller,
            args=(pair, poll_time),
            daemon=True
        )
        thread.start()

    # TODO: ideally this should use the RubiconMarket events to update itself instead of repeatedly polling the
    #  get_orderbook method. But it's fine for now.
    def _start_orderbook_poller(self, pair: Pair, poll_time: int = 2) -> None:
        """The internal implementation of the order book poller. It continuously retrieves the order book
        for the specified pair and adds it to the pair order books dictionary and the message queue of the client. The
        poller will run until the pair is removed from the client.

        :param pair: The pair to start the order book poller for.
        :type pair: Pair
        :param poll_time: Polling interval in seconds, defaults to 2 seconds.
        :type poll_time: int, optional
        """
        polling: bool = True
        while polling:
            try:
                order_book = self.get_orderbook(pair_name=pair.name)

                self._pair_orderbooks[pair.name] = order_book

                self.message_queue.put(order_book)
            except PairDoesNotExistException:
                log.warning("pair does not exist in client. shutting down orderbook poller")
                polling = False
            except Exception as e:
                log.error(e)
            sleep(poll_time)

    ######################################################################
    # event methods
    ######################################################################

    def start_event_poller(
        self,
        pair_name: str,
        event_type: Type[BaseEvent],
        filters: Optional[Dict[str, Any]] = None,
        event_handler: Optional[Callable] = None,
        poll_time: int = 2
    ) -> None:
        """Starts a background event poller that continuously listens for events of the specified event type
        related to the specified pair. The retrieved events are processed by the event handler and added to the message
        queue of the client. The poller will run until the pair is removed from the client.

        :param pair_name: Name of the pair to start the event poller for.
        :type pair_name: str
        :param event_type: Type of the event to listen for.
        :type event_type: Type[BaseEvent]
        :param filters: Optional filters to apply when retrieving events, defaults to the events default filters
            (optional, default is None).
        :type filters: Optional[Dict[str, Any]], optional
        :param event_handler: Optional event handler function to process the retrieved events, defaults to the
            self._default_event_handler (optional, default is None).
        :type event_handler: Optional[Callable], optional
        :param poll_time: Polling interval in seconds, defaults to 2 seconds.
        :type poll_time: int, optional
        :raises Exception: If the message queue is not configured.
        :raises PairDoesNotExistException: If the pair does not exist in the client.
        """
        if self.message_queue is None:
            raise Exception("event poller is configured to place messages on the message queue. message queue"
                            "cannot be none")

        pair = self.get_pair(pair_name)

        argument_filters = event_type.default_filters(
            bid_identifier=pair.bid_identifier,
            ask_identifier=pair.ask_identifier,
            wallet=self.wallet
        ) if filters is None else filters

        event_type.get_event_contract(market=self.market, router=self.router).start_event_poller(
            pair_name=pair_name,
            event_type=event_type,
            argument_filters=argument_filters,
            event_handler=self._default_event_handler if event_handler is None else event_handler,
            poll_time=poll_time
        )

    def _default_event_handler(self, pair_name: str, event_type: Type[BaseEvent], event_data: EventData) -> None:
        """The default event handler function used by the event poller. It processes the retrieved events
        and adds the corresponding order events to the message queue of the client.

        :param pair_name: Name of the pair associated with the event.
        :type pair_name: str
        :param event_type: Type of the event.
        :type event_type: Type[BaseEvent]
        :param event_data: Data of the retrieved event.
        :type event_data: EventData
        """
        raw_event = event_type(block_number=event_data["blockNumber"], **event_data["args"])

        if raw_event.client_filter(wallet=self.wallet):
            pair = self._pairs.get(pair_name)

            order_event = OrderEvent.from_event(pair=pair, event=raw_event, wallet=self.wallet)

            self.message_queue.put(order_event)

    ######################################################################
    # order methods
    ######################################################################
    # TODO: would be cool if these methods could understand how much they are spending on gas (use TxReceipt)
    #  also need a way to return the order id for limit orders
    #  we could listen for the event and return the order id along with relevant gas spend information 
    #  one thing we will need to be conscious of and figure out how to handle is the fact that gas is calculated in a
    #  variety of ways depending upon the chain

    def place_market_order(self, transaction: Transaction) -> str:
        """Place a market order transaction by executing the specified transaction object. The transaction
        object should contain a single order of type NewMarketOrder. The order is retrieved from the transaction and
        the corresponding market buy or sell method is called based on the order side.

        :param transaction: Transaction object containing the market order.
        :type transaction: Transaction
        :return: The transaction hash of the executed market order.
        :rtype: str
        :raises Exception: If the transaction contains more than one order.
        """
        if len(transaction.orders) > 1:
            raise Exception("call place_order with one order only")

        order: NewMarketOrder = transaction.orders[0]  # noqa

        pair = self.get_pair(pair_name=order.pair)

        match order.order_side:
            case OrderSide.BUY:
                return self.market.buy_all_amount(
                    buy_gem=pair.base_asset.address,
                    buy_amt=pair.base_asset.to_integer(order.size),
                    pay_gem=pair.quote_asset.address,
                    max_fill_amount=pair.quote_asset.to_integer(order.worst_execution_price),
                    nonce=transaction.nonce,
                    gas=transaction.gas,
                    max_fee_per_gas=transaction.max_fee_per_gas,
                    max_priority_fee_per_gas=transaction.max_priority_fee_per_gas
                )
            case OrderSide.SELL:
                return self.market.sell_all_amount(
                    pay_gem=pair.base_asset.address,
                    pay_amt=pair.base_asset.to_integer(order.size),
                    buy_gem=pair.quote_asset.address,
                    min_fill_amount=pair.quote_asset.to_integer(order.worst_execution_price),
                    nonce=transaction.nonce,
                    gas=transaction.gas,
                    max_fee_per_gas=transaction.max_fee_per_gas,
                    max_priority_fee_per_gas=transaction.max_priority_fee_per_gas
                )

    def place_limit_order(self, transaction: Transaction) -> str:
        """Place a limit order transaction by executing the specified transaction object. The transaction object should
        contain a single order of type NewLimitOrder.

        :param transaction: Transaction object containing the limit order.
        :type transaction: Transaction
        :return: The transaction hash of the executed limit order.
        :rtype: str
        :raises Exception: If the transaction contains more than one order.
        """
        if len(transaction.orders) > 1:
            raise Exception("call place_order with one order only")

        order: NewLimitOrder = transaction.orders[0]  # noqa

        pair = self.get_pair(pair_name=order.pair)

        match order.order_side:
            case OrderSide.BUY:
                return self.market.offer(
                    pay_amt=pair.quote_asset.to_integer(order.price * order.size),
                    pay_gem=pair.quote_asset.address,
                    buy_amt=pair.base_asset.to_integer(order.size),
                    buy_gem=pair.base_asset.address,
                    nonce=transaction.nonce,
                    gas=transaction.gas,
                    max_fee_per_gas=transaction.max_fee_per_gas,
                    max_priority_fee_per_gas=transaction.max_priority_fee_per_gas
                )
            case OrderSide.SELL:
                return self.market.offer(
                    pay_amt=pair.base_asset.to_integer(order.size),
                    pay_gem=pair.base_asset.address,
                    buy_amt=pair.quote_asset.to_integer(order.price * order.size),
                    buy_gem=pair.quote_asset.address,
                    nonce=transaction.nonce,
                    gas=transaction.gas,
                    max_fee_per_gas=transaction.max_fee_per_gas,
                    max_priority_fee_per_gas=transaction.max_priority_fee_per_gas
                )

    def cancel_limit_order(self, transaction: Transaction) -> str:
        """Place a limit order cancel transaction by executing the specified transaction object. The transaction object
        should contain a single order of type NewCancelOrder.

        :param transaction: Transaction object containing the cancel order.
        :type transaction: Transaction
        :return: The transaction hash of the executed cancel order.
        :rtype: str
        :raises Exception: If the transaction contains more than one order.
        """
        if len(transaction.orders) > 1:
            raise Exception("call place_order with one order only")

        order: NewCancelOrder = transaction.orders[0]  # noqa

        return self.market.cancel(
            id=order.order_id,
            nonce=transaction.nonce,
            gas=transaction.gas,
            max_fee_per_gas=transaction.max_fee_per_gas,
            max_priority_fee_per_gas=transaction.max_priority_fee_per_gas
        )

    def batch_place_limit_orders(self, transaction: Transaction) -> str:
        """Place multiple limit orders in a batch transaction.

        :param transaction: Transaction object containing multiple limit orders.
        :type transaction: Transaction
        :return: The transaction hash of the executed batch limit orders.
        :rtype: str
        """
        pay_amts = []
        pay_gems = []
        buy_amts = []
        buy_gems = []

        for order in transaction.orders:
            order: NewLimitOrder
            pair = self.get_pair(order.pair)

            match order.order_side:
                case OrderSide.BUY:
                    pay_amts.append(pair.quote_asset.to_integer(order.price * order.size))
                    pay_gems.append(pair.quote_asset.address)
                    buy_amts.append(pair.base_asset.to_integer(order.size))
                    buy_gems.append(pair.base_asset.address)
                case OrderSide.SELL:
                    pay_amts.append(pair.base_asset.to_integer(order.size))
                    pay_gems.append(pair.base_asset.address)
                    buy_amts.append(pair.quote_asset.to_integer(order.price * order.size))
                    buy_gems.append(pair.quote_asset.address)

        return self.market.batch_offer(
            pay_amts=pay_amts,
            pay_gems=pay_gems,
            buy_amts=buy_amts,
            buy_gems=buy_gems,
            nonce=transaction.nonce,
            gas=transaction.gas,
            max_fee_per_gas=transaction.max_fee_per_gas,
            max_priority_fee_per_gas=transaction.max_priority_fee_per_gas
        )

    def batch_update_limit_orders(self, transaction: Transaction) -> str:
        """Update multiple limit orders in a batch transaction.

        :param transaction: Transaction object containing multiple limit order updates.
        :type transaction: Transaction
        :return: The transaction hash of the executed batch limit order updates.
        :rtype: str
        """
        order_ids = []
        pay_amts = []
        pay_gems = []
        buy_amts = []
        buy_gems = []

        for order in transaction.orders:
            order: UpdateLimitOrder
            pair = self.get_pair(order.pair)

            order_ids.append(order.order_id)

            match order.order_side:
                case OrderSide.BUY:
                    pay_amts.append(pair.quote_asset.to_integer(order.price * order.size))
                    pay_gems.append(pair.quote_asset.address)
                    buy_amts.append(pair.base_asset.to_integer(order.size))
                    buy_gems.append(pair.base_asset.address)
                case OrderSide.SELL:
                    pay_amts.append(pair.base_asset.to_integer(order.size))
                    pay_gems.append(pair.base_asset.address)
                    buy_amts.append(pair.quote_asset.to_integer(order.price * order.size))
                    buy_gems.append(pair.quote_asset.address)

        return self.market.batch_offer(
            pay_amts=pay_amts,
            pay_gems=pay_gems,
            buy_amts=buy_amts,
            buy_gems=buy_gems,
            nonce=transaction.nonce,
            gas=transaction.gas,
            max_fee_per_gas=transaction.max_fee_per_gas,
            max_priority_fee_per_gas=transaction.max_priority_fee_per_gas
        )

    def batch_cancel_limit_orders(self, transaction: Transaction) -> str:
        """Cancel multiple limit orders in a batch transaction.

        :param transaction: Transaction object containing multiple limit order cancellations.
        :type transaction: Transaction
        :return: The transaction hash of the executed batch limit order cancellations.
        :rtype: str
        """
        order_ids = []

        for order in transaction.orders:
            order: NewCancelOrder

            order_ids.append(order.order_id)

        return self.market.batch_cancel(
            ids=order_ids,
            nonce=transaction.nonce,
            gas=transaction.gas,
            max_fee_per_gas=transaction.max_fee_per_gas,
            max_priority_fee_per_gas=transaction.max_priority_fee_per_gas
        )

    ######################################################################
    # helper methods
    ######################################################################

    # TODO: revisit as the safer thing is to set approval to 0 and then set approval to new_allowance
    #  or use increaseAllowance and decreaseAllowance but the current abi does not support these methods
    #  See: https://github.com/ethereum/EIPs/issues/20#issuecomment-263524729
    @staticmethod
    def _update_asset_allowance(
        asset: ERC20,
        spender: ChecksumAddress,
        new_allowance: Decimal
    ) -> None:
        log.info(
            asset.approve(
                spender=spender,
                amount=asset.to_integer(number=new_allowance)
            )
        )

    # TODO: implement this and use check transactions before they go through to prevent failure
    @staticmethod
    def _check_allowance(pair: Pair, order: BaseNewOrder):
        pass



from _decimal import Decimal
from typing import List, Tuple

from rubi import ERC20
from rubi.types.order import OrderSide


class BookLevel:
    """Class representing a level in the order book.

    :param price: The price of the level.
    :type price: Decimal
    :param size: The size of the level.
    :type size: Decimal
    """

    def __init__(
        self,
        price: Decimal,
        size: Decimal
    ):
        """constructor method."""
        self.price = price
        self.size = size

    def __repr__(self):
        items = ("{}={!r}".format(k, self.__dict__[k]) for k in self.__dict__)
        return "{}({})".format(type(self).__name__, ", ".join(items))


class BookSide:
    """Class Representing a side of the order book. Either bids or asks.

    :param book_side: The side of the order book (BUY or SELL).
    :type book_side: OrderSide
    :param levels: The list of levels on the side.
    :type levels: List[BookLevel]
    """

    def __init__(
        self,
        book_side: OrderSide,
        levels: List[BookLevel]
    ):
        """constructor method."""
        self.book_side = book_side
        self.levels = levels

    def best_price(self) -> Decimal:
        """Returns the price of the best level on the book side.

        :return: The price of the best level.
        :rtype: Decimal
        """
        return self.levels[0].price

    @classmethod
    def from_rubicon_offers(
        cls,
        book_side: OrderSide,
        offers: List[List[int]],
        base_asset: ERC20,
        quote_asset: ERC20
    ) -> "BookSide":
        """Creates a BookSide instance from a list of Rubicon offers.

        :param book_side: The side of the order book (BUY or SELL).
        :type book_side: OrderSide
        :param offers: The list of offers retrieved from the Rubicon for an asset pair pay_gem/buy_gem.
        :type offers: List[List[int]]
        :param base_asset: The base asset of the order book.
        :type base_asset: ERC20
        :param quote_asset: The quote asset of the order book.
        :type quote_asset: ERC20
        :return: The BookSide instance representing the order book side.
        :rtype: BookSide
        """
        levels: List[BookLevel] = []

        match book_side:
            case OrderSide.SELL:
                for i, order in enumerate(offers):
                    size = base_asset.to_decimal(order[0])
                    price = quote_asset.to_decimal(order[1]) / size

                    if levels and levels[-i].price == price:
                        levels[-1].size += size
                    else:
                        levels.append(BookLevel(price=price, size=size))
            case OrderSide.BUY:
                for i, order in enumerate(offers):
                    size = base_asset.to_decimal(order[1])
                    price = quote_asset.to_decimal(order[0]) / size

                    if levels and levels[-1].price == price:
                        levels[-1].size += size
                    else:
                        levels.append(BookLevel(price=price, size=size))

        return cls(
            book_side=book_side,
            levels=levels
        )

    def __repr__(self):
        items = ("{}={!r}".format(k, self.__dict__[k]) for k in self.__dict__)
        return "{}({})".format(type(self).__name__, ", ".join(items))


class OrderBook:
    """Class represents an OrderBook.

    :param bids: BookSide representing the bid orders.
    :type bids: BookSide
    :param asks: BookSide representing the ask orders.
    :type asks: BookSide
    """

    def __init__(self, bids: BookSide, asks: BookSide):
        """constructor method."""
        self.bids = bids
        self.asks = asks

    @classmethod
    def from_rubicon_offer_book(
        cls,
        offer_book: Tuple[List[List[int]], List[List[int]]],
        base_asset: ERC20,
        quote_asset: ERC20
    ) -> "OrderBook":
        """Create an OrderBook from Rubicon offer book.

        :param offer_book: Rubicon offer book containing bid and ask offers.
        :type offer_book: Tuple[List[List[int]], List[List[int]]]
        :param base_asset: An ERC20 instance representing the base asset.
        :type base_asset: ERC20
        :param quote_asset: An ERC20 instance representing the quote asset.
        :type quote_asset: ERC20
        :return: OrderBook instance.
        :rtype: OrderBook
        """
        return cls(
            bids=BookSide.from_rubicon_offers(
                book_side=OrderSide.BUY,  # Corresponds to BIDS
                offers=offer_book[1],
                base_asset=base_asset,
                quote_asset=quote_asset
            ),
            asks=BookSide.from_rubicon_offers(
                book_side=OrderSide.SELL,  # Corresponds to ASKS
                offers=offer_book[0],
                base_asset=base_asset,
                quote_asset=quote_asset
            )
        )

    def best_bid(self) -> Decimal:
        """Get the best bid price from the order book.

        :return: Best bid price.
        :rtype: Decimal
        """
        return self.bids.best_price()

    def best_ask(self) -> Decimal:
        """Get the best ask price from the order book.

        :return: Best ask price.
        :rtype: Decimal
        """
        return self.asks.best_price()

    def mid_price(self) -> Decimal:
        """Calculate the mid-price of the order book.

        :return: mid-price.
        :rtype: Decimal
        """
        return (self.best_bid() + self.best_ask()) / 2

    def __repr__(self):
        items = ("{}={!r}".format(k, self.__dict__[k]) for k in self.__dict__)
        return "{}({})".format(type(self).__name__, ", ".join(items))



import asyncio
import requests 
from .price import Price

class Gas:
    """Gas helper class"""

    def __init__(self, w3):
        """constructor for the gas helper class
        
        :param w3: a web3 instance
        :type w3: Web3 object
        """
        self.w3 = w3
        self.chain_id = w3.eth.chain_id
        self.price = Price()

    def get_optimism_txn_gas_data(self, txn, priced=True): 
        """this function takes a transaction hash and returns a dictionary of gas data for the transaction
        
        :param txn: the transaction hash
        :type txn: str
        :return: a dictionary of gas data for the transaction
        :rtype: dict
        """

        # TODO: Issue #18: multicall
        txn_object = self.w3.eth.get_transaction(txn)
        txn_receipt = self.w3.eth.get_transaction_receipt(txn)

        txn_data = {}
        txn_data['l2_gas_price'] = txn_object['gasPrice']
        txn_data['l2_gas_used'] = txn_receipt['gasUsed']
        txn_data['l1_gas_used'] = int(txn_receipt['l1GasUsed'], 0)
        txn_data['l1_gas_price'] = int(txn_receipt['l1GasPrice'], 0)
        txn_data['l1_fee_scalar'] = float(txn_receipt['l1FeeScalar'])
        txn_data['l1_fee'] = int(txn_data['l1_gas_used'] * txn_data['l1_gas_price'] * txn_data['l1_fee_scalar'])
        txn_data['l2_fee'] = txn_data['l2_gas_used'] * txn_data['l2_gas_price']
        txn_data['total_fee'] = txn_data['l1_fee'] + txn_data['l2_fee']
        txn_data['l1_fee_eth'] = txn_data['l1_fee'] / 10**18
        txn_data['l2_fee_eth'] = txn_data['l2_fee'] / 10**18
        txn_data['total_fee_eth'] = txn_data['total_fee'] / 10**18

        # this is intended to handle optimism mainnet transactions, pre-bedrock
        if self.chain_id == 10:
            txn_data['timestamp'] = int(txn_object['l1Timestamp'], 0)
        if self.chain_id == 420:
            txn_block = self.w3.eth.get_block(txn_object['blockNumber'])
            txn_data['timestamp'] = txn_block['timestamp']

        # if priced, get the price of eth at the time of the transaction
        if priced: 
            eth_price = self.price.get_defi_llama_price(timestamp=txn_data['timestamp'])

            # check that a valid price was returned
            if not eth_price:     
                eth_price = self.price.get_coinbase_price(date=txn_data['timestamp'])

            # check that a valid price was returned
            if not eth_price: 
                return None
            
            # calculate the usd value of the fees
            txn_data['eth_price'] = eth_price['amount']
            txn_data['l1_fee_usd'] = txn_data['l1_fee_eth'] * txn_data['eth_price']
            txn_data['l2_fee_usd'] = txn_data['l2_fee_eth'] * txn_data['eth_price']
            txn_data['total_fee_usd'] = txn_data['total_fee_eth'] * txn_data['eth_price']

        return txn_data

    async def get_optimism_txns_gas_data(self, txns):
        
        try: 
            loop = asyncio.get_event_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            
        tasks = [loop.run_in_executor(None, self.get_optimism_txn_gas_data, txn, False) for txn in txns]
        results = await asyncio.gather(*tasks)
        return results

    def retrieve_optimism_txns_gas_data(self, txns):

        try: 
            loop = asyncio.get_event_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        txns_data = loop.run_until_complete(self.get_optimism_txns_gas_data(txns))
        #loop.close()

        txns_data = dict(zip(txns, txns_data))
        return txns_data

    def get_closest_timestamp_value(self, values, timestamp, granularity=60, n=10): 
        # TODO: this is actually a function that is also included in data/processing/helper/processing.py and we should find a good way to consolidate these functions
        """this function takes a dictionary of values (assumes the keys are timestamps) and a timestamp and returns the value of the dictionary that is closest to the timestamp.

        :param values: a dictionary of values with the keys being timestamps
        :type values: dict
        :param timestamp: the timestamp that is of interest
        :type timestamp: int
        :param granularity: the granularity of the timestamps of interest, defaults to 60
        :type granularity: int, optional
        :param n: the optimized search default based upon granularity of which to search , defaults to 10
        :type n: int, optional
        :return: the value of the dictionary that is closest to the timestamp key pair that matches
        :rtype: value
        """

        # based on the granularity of the time perids, get the starting period value 
        timestamp = (timestamp // granularity) * granularity

        # get the keys of the dictionary 
        keys = list(values.keys())

        try: 
            value = values[timestamp]
            return value
        except:
            pass
        
        attempt = 1
        while attempt < n: 
            try: 
                value = values[timestamp + granularity * attempt]
                return value
            except:
                pass

            try:
                value = values[timestamp - granularity * attempt]
                return value
            except:
                pass

            attempt += 1
        
        # if this still doesn't work, search the entire list of keys
        min = abs(keys[0] - timestamp)
        value = values[keys[0]]
        for key in keys[1:]:
            if abs(key - timestamp) < min:
                min = abs(key - timestamp)
                value = values[key]

        return value

    def txn_dataframe_update(self, txn_dataframe, txn_column, timestamp_column, price_query='range', total_fee_eth = True, total_fee_usd = True, l2_gas_price = False, l2_gas_used = False, l1_gas_used = False, l1_gas_price = False, l1_fee_scalar = False, l1_fee = False, l2_fee = False, total_fee = False, l1_fee_eth = False, l2_fee_eth = False, eth_price = False, l1_fee_usd = False, l2_fee_usd = False): 
        """this function takes a dataframe of transactions and adds gas data to it. by default, it only adds the total gas fee in eth and usd to the dataframe. if any of the other parameters are set to true, it will add those values to the dataframe as well
        
        :param txn_dataframe: a dataframe of transactions
        :type txn_dataframe: pandas dataframe
        :param txn_column: the name of the column that contains the transaction hashes
        :type txn_column: str
        :param timestamp_column: the name of the column that contains the timestamp of the transaction
        :type timestamp_column: str
        :param price_query: the method to use to query the price of eth. can be 'range' or 'single'
        :type price_query: str
        :param total_fee_eth: whether or not to add the total fee in eth to the dataframe
        :type total_fee_eth: bool
        :param total_fee_usd: whether or not to add the total fee in usd to the dataframe
        :type total_fee_usd: bool
        :param l2_gas_price: whether or not to add the l2 gas price to the dataframe
        :type l2_gas_price: bool
        :param l2_gas_used: whether or not to add the l2 gas used to the dataframe
        :type l2_gas_used: bool
        :param l1_gas_used: whether or not to add the l1 gas used to the dataframe
        :type l1_gas_used: bool
        :param l1_gas_price: whether or not to add the l1 gas price to the dataframe
        :type l1_gas_price: bool
        :param l1_fee_scalar: whether or not to add the l1 fee scalar to the dataframe
        :type l1_fee_scalar: bool
        :param l1_fee: whether or not to add the l1 fee to the dataframe
        :type l1_fee: bool
        :param l2_fee: whether or not to add the l2 fee to the dataframe
        :type l2_fee: bool
        :param total_fee: whether or not to add the total fee to the dataframe
        :type total_fee: bool
        :param l1_fee_eth: whether or not to add the l1 fee in eth to the dataframe
        :type l1_fee_eth: bool
        :param l2_fee_eth: whether or not to add the l2 fee in eth to the dataframe
        :type l2_fee_eth: bool
        :param eth_price: whether or not to add the eth price to the dataframe
        :type eth_price: bool
        :param l1_fee_usd: whether or not to add the l1 fee in usd to the dataframe
        :type l1_fee_usd: bool
        :param l2_fee_usd: whether or not to add the l2 fee in usd to the dataframe
        :type l2_fee_usd: bool
        :return: a dataframe of transactions with gas data added
        :rtype: pandas dataframe
        """

        # get the unique transaction hashes from the dataframe
        txn_hashes = list(txn_dataframe[txn_column].unique())

        # from the transaction hashes and create a dictionary that maps the transaction the associated gas data
        txn_gas_data = self.retrieve_optimism_txns_gas_data(txn_hashes)

        if price_query == 'range':
            start_timestamp = txn_dataframe[timestamp_column].min()
            end_timestamp = txn_dataframe[timestamp_column].max()
            eth_prices = self.price.get_price_in_range(start_timestamp, end_timestamp)

            # create a dictionary that maps the timestamp to the associated eth price
            timestamps = list(txn_dataframe[timestamp_column].unique())
            eth_price_data = {}
            
            for timestamp in timestamps:
                eth_price_data[timestamp] = self.get_closest_timestamp_value(eth_prices, timestamp)

        else:
            # get the unique timestamps from the dataframe
            timestamps = list(txn_dataframe[timestamp_column].unique())

            # create an array of "ETH-USD" strings that represent the pair, the array is the same length as the timestamps array
            pairs = ['ETH-USD'] * len(timestamps)
            granularities = [60] * len(timestamps)

            # from the timestamps and create a dictionary that maps the timestamp to the associated eth price
            eth_prices = self.price.retrieve_prices(pairs, granularities, timestamps)
            eth_price_data = dict(zip(timestamps, eth_prices))

        # now add all columns that we are interested in to the dataframe
        if total_fee_eth:
            txn_dataframe['total_fee_eth'] = txn_dataframe[txn_column].map(lambda x: txn_gas_data[x]['total_fee_eth'])
        if total_fee_usd: 
            txn_dataframe['total_fee_usd'] = txn_dataframe.apply(lambda x: txn_gas_data[x[txn_column]]['total_fee_eth'] * eth_price_data[x[timestamp_column]], axis=1)
        if l2_gas_price:
            txn_dataframe['l2_gas_price'] = txn_dataframe[txn_column].map(lambda x: txn_gas_data[x]['l2_gas_price'])
        if l2_gas_used:
            txn_dataframe['l2_gas_used'] = txn_dataframe[txn_column].map(lambda x: txn_gas_data[x]['l2_gas_used'])
        if l1_gas_used:
            txn_dataframe['l1_gas_used'] = txn_dataframe[txn_column].map(lambda x: txn_gas_data[x]['l1_gas_used'])
        if l1_gas_price:
            txn_dataframe['l1_gas_price'] = txn_dataframe[txn_column].map(lambda x: txn_gas_data[x]['l1_gas_price'])
        if l1_fee_scalar:
            txn_dataframe['l1_fee_scalar'] = txn_dataframe[txn_column].map(lambda x: txn_gas_data[x]['l1_fee_scalar'])
        if l1_fee:
            txn_dataframe['l1_fee'] = txn_dataframe[txn_column].map(lambda x: txn_gas_data[x]['l1_fee'])
        if l2_fee:
            txn_dataframe['l2_fee'] = txn_dataframe[txn_column].map(lambda x: txn_gas_data[x]['l2_fee'])
        if total_fee:
            txn_dataframe['total_fee'] = txn_dataframe[txn_column].map(lambda x: txn_gas_data[x]['total_fee'])
        if l1_fee_eth:
            txn_dataframe['l1_fee_eth'] = txn_dataframe[txn_column].map(lambda x: txn_gas_data[x]['l1_fee_eth'])
        if l2_fee_eth:
            txn_dataframe['l2_fee_eth'] = txn_dataframe[txn_column].map(lambda x: txn_gas_data[x]['l2_fee_eth'])
        if eth_price:
            txn_dataframe['eth_price'] = txn_dataframe.map(lambda x: eth_price_data[x[timestamp_column]])
        if l1_fee_usd:
            txn_dataframe['l1_fee_usd'] = txn_dataframe.map(lambda x: txn_gas_data[x[txn_column]]['l1_fee_eth'] * eth_price_data[x[timestamp_column]])
        if l2_fee_usd:
            txn_dataframe['l2_fee_usd'] = txn_dataframe.map(lambda x: txn_gas_data[x[txn_column]]['l2_fee_eth'] * eth_price_data[x[timestamp_column]])
        
        return txn_dataframe
